@@ -9,6 +9,7 @@ void log_data(unsigned char *data, int len)
 {
 	int fd;
 	if ((fd = open(LOG_TRANSFER, O_WRONLY | O_APPEND | O_CREAT, 0622)) != -1) {
+		set_bin(fd);
 		write(fd, data, len);
 		close(fd);
 	}
@@ -46,6 +47,7 @@ void dns_found(struct connection *, int);
 
 void make_connection(struct connection *c, int port, int *sock, void (*func)(struct connection *))
 {
+	int as;
 	unsigned char *host;
 	struct conn_info *b;
 	if (!(host = get_host_name(c->url))) {
@@ -63,10 +65,13 @@ void make_connection(struct connection *c, int port, int *sock, void (*func)(str
 	b->sock = sock;
 	b->port = port;
 	c->buffer = b;
-	setcstate(c, S_DNS);
-	if (c->no_cache >= NC_RELOAD) find_host_no_cache(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
-	else find_host(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
+	log_data("\nCONNECTION: ", 13);
+	log_data(host, strlen(host));
+	log_data("\n", 1);
+	if (c->no_cache >= NC_RELOAD) as = find_host_no_cache(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
+	else as = find_host(host, &b->addr, &c->dnsquery, (void(*)(void *, int))dns_found, c);
 	mem_free(host);
+	if (as) setcstate(c, S_DNS);
 }
 
 int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port)
@@ -84,7 +89,6 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) goto e;
 	*sock = s;
 	fcntl(s, F_SETFL, O_NONBLOCK);
-	again:
 	memcpy(&sb, &sa, sizeof(struct sockaddr_in));
 	sb.sin_port = 0;
 	if (bind(s, (struct sockaddr *)&sb, sizeof sb)) goto e;
@@ -93,7 +97,7 @@ int get_pasv_socket(struct connection *c, int cc, int *sock, unsigned char *port
 	if (listen(s, 1)) goto e;
 	memcpy(port, &sa.sin_addr.s_addr, 4);
 	memcpy(port + 4, &sa.sin_port, 2);
-#ifdef IP_TOS
+#if defined(IP_TOS) && defined(IPTOS_THROUGHPUT)
 	{
 		int on = IPTOS_THROUGHPUT;
 		setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int));
@@ -111,7 +115,6 @@ void dns_found(struct connection *c, int state)
 		retry_connection(c);
 		return;
 	}
-	setcstate(c, S_CONN);
 	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
 		setcstate(c, -errno);
 		retry_connection(c);
@@ -130,6 +133,7 @@ void dns_found(struct connection *c, int state)
 			return;
 		}
 		set_handlers(s, NULL, (void(*)(void *))connected, (void(*)(void *))exception, c);
+		setcstate(c, S_CONN);
 	} else {
 		c->buffer = NULL;
 		b->func(c);
@@ -141,15 +145,26 @@ void connected(struct connection *c)
 {
 	struct conn_info *b = c->buffer;
 	void (*func)(struct connection *) = b->func;
+	int err = 0;
+	int len = sizeof(int);
 	c->buffer = NULL;
-	if (!connect(*b->sock, (struct sockaddr *)&b->sa, sizeof b->sa) || errno == EISCONN) {
+	/*if (!connect(*b->sock, (struct sockaddr *)&b->sa, sizeof b->sa) || errno == EISCONN) {
 		mem_free(b);
 		func(c);
 		return;
 	}
 	mem_free(b);
-	setcstate(c, -errno);
-	retry_connection(c);
+	setcstate(c, -errno);*/
+	if (getsockopt(*b->sock, SOL_SOCKET, SO_ERROR, (void *)&err, &len))
+		if (!(err = errno)) {
+			err = -S_STATE;
+			goto bla;
+		}
+	if (err >= 10000) err -= 10000;	/* Why does EMX return so large values? */
+	bla:
+	mem_free(b);
+	if (err > 0) setcstate(c, -err), retry_connection(c);
+	else func(c);
 }
 
 struct write_buffer {
