@@ -13,10 +13,10 @@ int parse_element(unsigned char *e, unsigned char *eof, unsigned char **name, in
 	if (name) *name = e;
 	if (e < eof && *e == '/') e++;
 	while (e < eof && isA(*e)) e++;
-	if (e >= eof || (!WHITECHAR(*e) && *e != '>')) return -1;
+	if (e >= eof || (!WHITECHAR(*e) && *e != '>' && *e != '/' && *e != ':')) return -1;
 	if (name && namelen) *namelen = e - *name;
-	while (e < eof && WHITECHAR(*e)) e++;
-	if (e >= eof || (!isA(*e) && *e != '>')) return -1;
+	while (e < eof && (WHITECHAR(*e) || *e == '/') || *e == ':') e++;
+	if (e >= eof || (!atchr(*e) && *e != '>')) return -1;
 	if (attr) *attr = e;
 	nextattr:
 	while (e < eof && WHITECHAR(*e)) e++;
@@ -122,6 +122,19 @@ int has_attr(unsigned char *e, unsigned char *name)
 	return 1;
 }
 
+unsigned char *get_url_val(unsigned char *e, unsigned char *name)
+{
+	unsigned char *p, *q, *pp;
+	if (!(pp = get_attr_val(e, name))) return NULL;
+	p = pp; q = pp;
+	while (1) {
+		if ((*p = *q) != ' ') p++;
+		if (!*q) break;
+		q++;
+	}
+	return pp;
+}
+
 struct {
 	int n;
 	char *s;
@@ -181,44 +194,172 @@ void roman(char *p, unsigned n)
 
 struct color_spec {
 	char *name;
-	char *rgb;
+	int rgb;
 };
 
 struct color_spec color_specs[] = {
-	{"black",	"#000000"},
-	{"gray",	"#808080"},
-	{"silver",	"#c0c0c0"},
-	{"white",	"#ffffff"},
-	{"maroon",	"#800000"},
-	{"red",		"#ff0000"},
-	{"purple",	"#800080"},
-	{"fuchsia",	"#ff00ff"},
-	{"green",	"#008000"},
-	{"lime",	"#00ff00"},
-	{"olive",	"#808000"},
-	{"yellow",	"#ffff00"},
-	{"navy",	"#000080"},
-	{"blue",	"#0000ff"},
-	{"teal",	"#008080"},
-	{"aqua",	"#00ffff"},
-	{NULL,		NULL}
+	{"aliceblue",		0xF0F8FF},
+	{"antiquewhite",	0xFAEBD7},
+	{"aqua",		0x00FFFF},
+	{"aquamarine",		0x7FFFD4},
+	{"azure",		0xF0FFFF},
+	{"beige",		0xF5F5DC},
+	{"bisque",		0xFFE4C4},
+	{"black",		0x000000},
+	{"blanchedalmond",	0xFFEBCD},
+	{"blue",		0x0000FF},
+	{"blueviolet",		0x8A2BE2},
+	{"brown",		0xA52A2A},
+	{"burlywood",		0xDEB887},
+	{"cadetblue",		0x5F9EA0},
+	{"chartreuse",		0x7FFF00},
+	{"chocolate",		0xD2691E},
+	{"coral",		0xFF7F50},
+	{"cornflowerblue",	0x6495ED},
+	{"cornsilk",		0xFFF8DC},
+	{"crimson",		0xDC143C},
+	{"cyan",		0x00FFFF},
+	{"darkblue",		0x00008B},
+	{"darkcyan",		0x008B8B},
+	{"darkgoldenrod",	0xB8860B},
+	{"darkgray",		0xA9A9A9},
+	{"darkgreen",		0x006400},
+	{"darkkhaki",		0xBDB76B},
+	{"darkmagenta",		0x8B008B},
+	{"darkolivegreen",	0x556B2F},
+	{"darkorange",		0xFF8C00},
+	{"darkorchid",		0x9932CC},
+	{"darkred",		0x8B0000},
+	{"darksalmon",		0xE9967A},
+	{"darkseagreen",	0x8FBC8F},
+	{"darkslateblue",	0x483D8B},
+	{"darkslategray",	0x2F4F4F},
+	{"darkturquoise",	0x00CED1},
+	{"darkviolet",		0x9400D3},
+	{"deeppink",		0xFF1493},
+	{"deepskyblue",		0x00BFFF},
+	{"dimgray",		0x696969},
+	{"dodgerblue",		0x1E90FF},
+	{"firebrick",		0xB22222},
+	{"floralwhite",		0xFFFAF0},
+	{"forestgreen",		0x228B22},
+	{"fuchsia",		0xFF00FF},
+	{"gainsboro",		0xDCDCDC},
+	{"ghostwhite",		0xF8F8FF},
+	{"gold",		0xFFD700},
+	{"goldenrod",		0xDAA520},
+	{"gray",		0x808080},
+	{"green",		0x008000},
+	{"greenyellow",		0xADFF2F},
+	{"honeydew",		0xF0FFF0},
+	{"hotpink",		0xFF69B4},
+	{"indianred",		0xCD5C5C},
+	{"indigo",		0x4B0082},
+	{"ivory",		0xFFFFF0},
+	{"khaki",		0xF0E68C},
+	{"lavender",		0xE6E6FA},
+	{"lavenderblush",	0xFFF0F5},
+	{"lawngreen",		0x7CFC00},
+	{"lemonchiffon",	0xFFFACD},
+	{"lightblue",		0xADD8E6},
+	{"lightcoral",		0xF08080},
+	{"lightcyan",		0xE0FFFF},
+	{"lightgoldenrodyellow",	0xFAFAD2},
+	{"lightgreen",		0x90EE90},
+	{"lightgrey",		0xD3D3D3},
+	{"lightpink",		0xFFB6C1},
+	{"lightsalmon",		0xFFA07A},
+	{"lightseagreen",	0x20B2AA},
+	{"lightskyblue",	0x87CEFA},
+	{"lightslategray",	0x778899},
+	{"lightsteelblue",	0xB0C4DE},
+	{"lightyellow",		0xFFFFE0},
+	{"lime",		0x00FF00},
+	{"limegreen",		0x32CD32},
+	{"linen",		0xFAF0E6},
+	{"magenta",		0xFF00FF},
+	{"maroon",		0x800000},
+	{"mediumaquamarine",	0x66CDAA},
+	{"mediumblue",		0x0000CD},
+	{"mediumorchid",	0xBA55D3},
+	{"mediumpurple",	0x9370DB},
+	{"mediumseagreen",	0x3CB371},
+	{"mediumslateblue",	0x7B68EE},
+	{"mediumspringgreen",	0x00FA9A},
+	{"mediumturquoise",	0x48D1CC},
+	{"mediumvioletred",	0xC71585},
+	{"midnightblue",	0x191970},
+	{"mintcream",		0xF5FFFA},
+	{"mistyrose",		0xFFE4E1},
+	{"moccasin",		0xFFE4B5},
+	{"navajowhite",		0xFFDEAD},
+	{"navy",		0x000080},
+	{"oldlace",		0xFDF5E6},
+	{"olive",		0x808000},
+	{"olivedrab",		0x6B8E23},
+	{"orange",		0xFFA500},
+	{"orangered",		0xFF4500},
+	{"orchid",		0xDA70D6},
+	{"palegoldenrod",	0xEEE8AA},
+	{"palegreen",		0x98FB98},
+	{"paleturquoise",	0xAFEEEE},
+	{"palevioletred",	0xDB7093},
+	{"papayawhip",		0xFFEFD5},
+	{"peachpuff",		0xFFDAB9},
+	{"peru",		0xCD853F},
+	{"pink",		0xFFC0CB},
+	{"plum",		0xDDA0DD},
+	{"powderblue",		0xB0E0E6},
+	{"purple",		0x800080},
+	{"red",			0xFF0000},
+	{"rosybrown",		0xBC8F8F},
+	{"royalblue",		0x4169E1},
+	{"saddlebrown",		0x8B4513},
+	{"salmon",		0xFA8072},
+	{"sandybrown",		0xF4A460},
+	{"seagreen",		0x2E8B57},
+	{"seashell",		0xFFF5EE},
+	{"sienna",		0xA0522D},
+	{"silver",		0xC0C0C0},
+	{"skyblue",		0x87CEEB},
+	{"slateblue",		0x6A5ACD},
+	{"slategray",		0x708090},
+	{"snow",		0xFFFAFA},
+	{"springgreen",		0x00FF7F},
+	{"steelblue",		0x4682B4},
+	{"tan",			0xD2B48C},
+	{"teal",		0x008080},
+	{"thistle",		0xD8BFD8},
+	{"tomato",		0xFF6347},
+	{"turquoise",		0x40E0D0},
+	{"violet",		0xEE82EE},
+	{"wheat",		0xF5DEB3},
+	{"white",		0xFFFFFF},
+	{"whitesmoke",		0xF5F5F5},
+	{"yellow",		0xFFFF00},
+	{"yellowgreen",		0x9ACD32},
 };
 
+#define endof(T) ((T)+sizeof(T)/sizeof(*(T)))
+
 int decode_color(unsigned char *str, struct rgb *col)
 {
+	int ch;
 	if (*str != '#') {
 		struct color_spec *cs;
-		for (cs = color_specs; cs->name; cs++) if (!strcasecmp(cs->name, str)) {
-			str = cs->rgb;
-			goto found;
-		}
-		return -1;
+		for (cs = color_specs; cs < endof(color_specs); cs++)
+			if (!strcasecmp(cs->name, str)) {
+				ch = cs->rgb;
+				goto found;
+			}
+		str--;
 	}
-	found:
-	if (strlen(str) == 7) {
+	str++;
+	if (strlen(str) == 6) {
 		char *end;
-		int ch = strtoul(str+1, &end, 16);
+		ch = strtoul(str, &end, 16);
 		if (!*end) {
+found:
 			col->r = ch / 0x10000;
 			col->g = ch / 0x100 % 0x100;
 			col->b = ch % 0x100;
@@ -245,6 +386,18 @@ int get_bgcolor(unsigned char *a, struct rgb *rgb)
 	return get_color(a, "bgcolor", rgb);
 }
 
+unsigned char *get_target(unsigned char *a)
+{
+	unsigned char *v = get_attr_val(a, "target");
+	if (v) {
+		if (!strcasecmp(v, "_self")) {
+			mem_free(v);
+			v = stracpy(d_opt->framename);
+		}
+	}
+	return v;
+}
+
 void kill_html_stack_item(struct html_element *e)
 {
 	if (e->dontkill == 2) {
@@ -375,7 +528,6 @@ void kill_until(int ls, ...)
 	struct html_element *e = &html_top;
 	if (ls) e = e->next;
 	while ((void *)e != &html_stack) {
-		int i;
 		int sk = 0;
 		va_list arg;
 		va_start(arg, ls);
@@ -416,47 +568,58 @@ int get_num(unsigned char *a, unsigned char *n)
 	if ((al = get_attr_val(a, n))) {
 		char *end;
 		int s = strtoul(al, &end, 10);
-		if (!*al || s < 0) s = -1;
+		if (!*al || *end || s < 0) s = -1;
 		mem_free(al);
 		return s;
 	}
 	return -1;
 }
 
-int get_width(unsigned char *a, unsigned char *n, int trunc)
+int parse_width(unsigned char *w, int trunc)
 {
-	char *w, *end;
+	unsigned char *end;
 	int p = 0;
 	int s;
-	if (!(w = get_attr_val(a, n))) return -1;
-	if (!*w) {
-		mem_free(w);
-		return -1;
-	}
-	if (w[strlen(w) - 1] == '%') w[strlen(w) - 1] = 0, p = 1;
-	s = strtoul(w, &end, 10);
-	if (*end) {
-		mem_free(w);
-		return -1;
-	}
+	int l;
+	while (WHITECHAR(*w)) w++;
+	for (l = 0; w[l] && w[l] != ','; l++) ;
+	while (l && WHITECHAR(w[l - 1])) l--;
+	if (!l) return -1;
+	if (w[l - 1] == '%') l--, p = 1;
+	while (l && WHITECHAR(w[l - 1])) l--;
+	if (!l) return -1;
+	s = strtoul((char *)w, (char **)&end, 10);
+	if (end - w < l) return -1;
 	if (p) {
 		if (trunc) s = s * (par_format.width - par_format.leftmargin - par_format.rightmargin) / 100;
-		else {
-			mem_free(w);
-			return -1;
-		}
+		else return -1;
 	} else s = (s + (HTML_CHAR_WIDTH - 1) / 2) / HTML_CHAR_WIDTH;
 	if (trunc && s > par_format.width - par_format.leftmargin - par_format.rightmargin) s = par_format.width - par_format.leftmargin - par_format.rightmargin;
 	if (s < 0) s = 0;
-	mem_free(w);
 	return s;
 }
 
-int form_num;
-struct form form = { 0, NULL, 0};
-int g_ctrl_num;
+int get_width(unsigned char *a, unsigned char *n, int trunc)
+{
+	int r;
+	unsigned char *w;
+	if (!(w = get_attr_val(a, n))) return -1;
+	r = parse_width(w, trunc);
+	mem_free(w);
+	return r;
+}
 
-void put_link_line(unsigned char *prefix, unsigned char *linkname, unsigned char *link)
+/*int form_num;
+struct form form = { 0, NULL, 0 };
+int g_ctrl_num;*/
+
+struct form form = { NULL, NULL, 0, 0 };
+
+unsigned char *last_form_tag;
+unsigned char *last_form_attr;
+unsigned char *last_input_tag;
+
+void put_link_line(unsigned char *prefix, unsigned char *linkname, unsigned char *link, unsigned char *target)
 {
 	html_stack_dup();
 	ln_break(1, line_break_f, ff);
@@ -465,7 +628,7 @@ void put_link_line(unsigned char *prefix, unsigned char *linkname, unsigned char
 	format.form = NULL;
 	put_chrs(prefix, strlen(prefix), put_chars_f, ff);
 	format.link = join_urls(format.href_base, link);
-	format.target = stracpy("");
+	format.target = stracpy(target);
 	memcpy(&format.fg, &format.clink, sizeof(struct rgb));
 	put_chrs(linkname, strlen(linkname), put_chars_f, ff);
 	ln_break(1, line_break_f, ff);
@@ -481,14 +644,14 @@ void html_fixed(unsigned char *a) { format.attr |= AT_FIXED; }
 void html_a(unsigned char *a)
 {
 	char *al;
-	if ((al = get_attr_val(a, "href"))) {
+	if ((al = get_url_val(a, "href"))) {
 		char *all = al;
 		while (all[0] == ' ') all++;
 		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
 		if (format.link) mem_free(format.link);
 		format.link = join_urls(format.href_base, all);
 		mem_free(al);
-		if ((al = get_attr_val(a, "target"))) {
+		if ((al = get_target(a))) {
 			if (format.target) mem_free(format.target);
 			format.target = al;
 		} else {
@@ -550,7 +713,7 @@ void html_img(unsigned char *a)
 	ismap = format.link && has_attr(a, "ismap") && !usemap;
 	if (!(al = get_attr_val(a, "alt")) || !*al) {
 		if (al) mem_free(al);
-		if (!display_images && !format.link) return;
+		if (!d_opt->images && !format.link) return;
 		if (usemap) al = stracpy("[USEMAP]");
 		else if (ismap) al = stracpy("[ISMAP]");
 		else al = stracpy("[IMG]");
@@ -558,7 +721,7 @@ void html_img(unsigned char *a)
 	if (format.image) mem_free(format.image), format.image = NULL;
 	if (al) {
 		unsigned char *s;
-		if ((s = get_attr_val(a, "src"))) {
+		if ((s = get_url_val(a, "src")) || (s = get_attr_val(a, "dynsrc"))) {
 			format.image = join_urls(format.href_base, s);
 			mem_free(s);
 		}
@@ -581,7 +744,6 @@ void html_img(unsigned char *a)
 
 void html_body(unsigned char *a)
 {
-	char *al;
 	get_color(a, "text", &format.fg);
 	get_color(a, "link", &format.clink);
 	get_color(a, "vlink", &format.vlink);
@@ -621,6 +783,11 @@ void html_br(unsigned char *a)
 	was_br = 1;
 }
 
+void html_form(unsigned char *a)
+{
+	was_br = 1;
+}
+
 void html_p(unsigned char *a)
 {
 	if (par_format.leftmargin < margin) par_format.leftmargin = margin;
@@ -680,7 +847,6 @@ void html_hr(unsigned char *a)
 {
 	int i/* = par_format.width - 10*/;
 	char r = 205;
-	char *al;
 	int q = get_num(a, "size");
 	if (q >= 0 && q < 2) r = 196;
 	html_stack_dup();
@@ -731,12 +897,12 @@ void html_td(unsigned char *a)
 void html_base(unsigned char *a)
 {
 	char *al;
-	if ((al = get_attr_val(a, "href"))) {
+	if ((al = get_url_val(a, "href"))) {
 		if (format.href_base) mem_free(format.href_base);
 		format.href_base = join_urls(((struct html_element *)html_stack.prev)->attr.href_base, al);
 		mem_free(al);
 	}
-	if ((al = get_attr_val(a, "target"))) {
+	if ((al = get_target(a))) {
 		if (format.target_base) mem_free(format.target_base);
 		format.target_base = al;
 	}
@@ -855,39 +1021,81 @@ void html_dd(unsigned char *a)
 	par_format.align = AL_LEFT;
 }
 
-void html_form(unsigned char *a)
+void get_html_form(unsigned char *a, struct form *form)
 {
 	char *al;
-	form_num++;
-	if (form.action) mem_free(form.action);
 	if ((al = get_attr_val(a, "action"))) {
 		char *all = al;
 		while (all[0] == ' ') all++;
 		while (all[0] && all[strlen(all) - 1] == ' ') all[strlen(all) - 1] = 0;
-		form.action = join_urls(format.href_base, all);
+		form->action = join_urls(format.href_base, all);
 		mem_free(al);
-	} else form.action = NULL;
-	form.method = FM_GET;
+	} else form->action = stracpy(format.href_base);
+	if ((al = get_target(a))) {
+		form->target = al;
+	} else {
+		form->target = stracpy(format.target_base);
+	}
+	form->method = FM_GET;
 	if ((al = get_attr_val(a, "method"))) {
-		if (!strcasecmp(al, "get")) form.method = FM_GET;
+		if (!strcasecmp(al, "get")) form->method = FM_GET;
 		if (!strcasecmp(al, "post")) {
 			char *ax;
-			form.method = FM_POST;
+			form->method = FM_POST;
 			if ((ax = get_attr_val(a, "enctype"))) {
 				if (!strcasecmp(ax, "multipart/form-data"))
-					form.method = FM_POST_MP;
+					form->method = FM_POST_MP;
 				mem_free(ax);
 			}
 		}
 		mem_free(al);
 	}
-	form.ctrl_num = 0;
+	form->num = a - startf;
+}
+
+void find_form_for_input(unsigned char *i)
+{
+	unsigned char *s, *ss, *name, *attr, *lf, *la;
+	int namelen;
+	if (form.action) mem_free(form.action);
+	if (form.target) mem_free(form.target);
+	if (last_input_tag && i <= last_input_tag && i > last_form_tag) {
+		get_html_form(last_form_attr, &form);
+		return;
+	}
+	if (last_input_tag && i > last_input_tag) s = last_form_tag;
+	else s = startf;
+	lf = NULL, la = NULL;
+	se:
+	while (s < i && *s != '<') sp:s++;
+	if (s >= i) goto end_parse;
+	if (s + 2 <= eofff && (s[1] == '!' || s[1] == '?')) {
+		s = skip_comment(s, i);
+		goto se;
+	}
+	ss = s;
+	if (parse_element(s, i, &name, &namelen, &attr, &s)) goto sp;
+	if (namelen != 4 || casecmp(name, "FORM", 4)) goto se;
+	lf = ss;
+	la = attr;
+	goto se;
+
+	end_parse:
+	if (lf) {
+		last_form_tag = lf;
+		last_form_attr = la;
+		last_input_tag = i;
+		get_html_form(la, &form);
+	} else {
+		memset(&form, 0, sizeof(struct form));
+	}
 }
 
 void html_button(unsigned char *a)
 {
 	char *al;
 	struct form_control *fc;
+	find_form_for_input(a);
 	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
 	memset(fc, 0, sizeof(struct form_control));
 	if (!(al = get_attr_val(a, "type"))) {
@@ -913,9 +1121,8 @@ void html_button(unsigned char *a)
 	}
 	mem_free(al);
 	xxx:
-	fc->form_num = form_num;
-	fc->ctrl_num = form.ctrl_num++;
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->form_num = last_form_tag - startf;
+	fc->ctrl_num = a - last_form_tag;
 	fc->position = a - startf;
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
@@ -940,6 +1147,7 @@ void html_input(unsigned char *a)
 	int i;
 	unsigned char *al;
 	struct form_control *fc;
+	find_form_for_input(a);
 	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
 	memset(fc, 0, sizeof(struct form_control));
 	if (!(al = get_attr_val(a, "type"))) {
@@ -968,14 +1176,15 @@ void html_input(unsigned char *a)
 	} else fc->type = FC_TEXT;
 	mem_free(al);
 	xxx:
-	fc->form_num = form_num;
-	fc->ctrl_num = form.ctrl_num++;
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->form_num = last_form_tag - startf;
+	fc->ctrl_num = a - last_form_tag;
 	fc->position = a - startf;
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
+	fc->target = stracpy(form.target);
 	fc->name = get_attr_val(a, "name");
 	if (fc->type != FC_FILE) fc->default_value = get_attr_val(a, "value");
+	if (fc->type == FC_CHECKBOX && !fc->default_value) fc->default_value = stracpy("on");
 	if ((fc->size = get_num(a, "size")) == -1) fc->size = HTML_DEFAULT_INPUT_SIZE;
 	if ((fc->maxlength = get_num(a, "maxlength")) == -1) fc->maxlength = MAXINT;
 	if (fc->type == FC_CHECKBOX || fc->type == FC_RADIO) fc->default_state = has_attr(a, "checked");
@@ -1002,7 +1211,7 @@ void html_input(unsigned char *a)
 			break;
 		case FC_IMAGE:
 			if (format.image) mem_free(format.image), format.image = NULL;
-			if ((al = get_attr_val(a, "src"))) {
+			if ((al = get_url_val(a, "src")) || (al = get_url_val(a, "dynsrc"))) {
 				format.image = join_urls(format.href_base, al);
 				mem_free(al);
 			}
@@ -1042,6 +1251,7 @@ void html_option(unsigned char *a)
 {
 	struct form_control *fc;
 	unsigned char *val;
+	find_form_for_input(a);
 	if (!format.select) return;
 	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
 	memset(fc, 0, sizeof(struct form_control));
@@ -1078,9 +1288,8 @@ void html_option(unsigned char *a)
 		    (namelen == 9 && !casecmp(name, "/OPTGROUP", 9)))) goto r;
 	}
 	x:
-	fc->form_num = form_num;
-	fc->ctrl_num = form.ctrl_num++;
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->form_num = last_form_tag - startf;
+	fc->ctrl_num = a - last_form_tag;
 	fc->position = a - startf;
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
@@ -1104,7 +1313,7 @@ void clr_spaces(unsigned char *name)
 {
 	unsigned char *nm;
 	for (nm = name; *nm; nm++)
-		if (WHITECHAR(*nm)) *nm = ' ';
+		if (WHITECHAR(*nm) || *nm == 1) *nm = ' ';
 	for (nm = name; *nm; nm++)
 		while (nm[0] == ' ' && (nm == name || nm[1] == ' ' || !nm[1]))
 			memmove(nm, nm + 1, strlen(nm));
@@ -1113,9 +1322,10 @@ void clr_spaces(unsigned char *name)
 int menu_stack_size;
 struct menu_item **menu_stack;
 
-void new_menu_item(unsigned char *name, int data) /* name == NULL - up;	data == -1 - down */
+void new_menu_item(unsigned char *name, int data, int fullname)
+	/* name == NULL - up;	data == -1 - down */
 {
-	struct menu_item *top, *item, *nmenu;
+	struct menu_item *top, *item, *nmenu = NULL; /* no uninitialized warnings */
 	if (name) {
 		clr_spaces(name);
 		if (!name[0]) mem_free(name), name = stracpy(" ");
@@ -1146,7 +1356,7 @@ void new_menu_item(unsigned char *name, int data) /* name == NULL - up;	data ==
 		}
 		item->text = name;
 		item->rtext = data == -1 ? ">" : "";
-		item->hotkey = 0;
+		item->hotkey = fullname ? "\000\001" : "\000\000"; /* dirty */
 		item->func = data == -1 ? MENU_FUNC do_select_submenu : MENU_FUNC selected_item;
 		item->data = data == -1 ? nmenu : (void *)data;
 		item->in_m = data == -1 ? 1 : 0;
@@ -1168,7 +1378,7 @@ void init_menu()
 {
 	menu_stack_size = 0;
 	menu_stack = DUMMY;
-	new_menu_item(stracpy(""), -1);
+	new_menu_item(stracpy(""), -1, 0);
 }
 
 void free_menu(struct menu_item *m) /* Grrr. Recursion */
@@ -1207,7 +1417,7 @@ void menu_labels(struct menu_item *m, unsigned char *base, unsigned char **lbls)
 				mem_free(bs);
 			}
 		} else {
-			if ((bs = stracpy(base))) add_to_strn(&bs, m->text);
+			if ((bs = stracpy(m->hotkey[1] ? (unsigned char *)"" : base))) add_to_strn(&bs, m->text);
 			lbls[(int)m->data] = bs;
 		}
 	}
@@ -1239,11 +1449,13 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	int t_namelen;
 	unsigned char *lbl;
 	int lbl_l;
+	int nnmi = 0;
 	struct conv_table *ct = special_f(f, SP_TABLE, NULL);
 	unsigned char **val, **lbls;
 	int order, preselect, group;
 	int i, mw;
 	if (has_attr(attr, "multiple")) return 1;
+	find_form_for_input(attr);
 	lbl = NULL;
 	val = DUMMY;
 	order = 0, group = 0, preselect = -1;
@@ -1269,7 +1481,7 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		int l = html - en;
 		while (l && WHITECHAR(s[0])) s++, l--;
 		while (l && WHITECHAR(s[l-1])) l--;
-		q = convert_string(special_f(f, SP_TABLE, NULL), s, l);
+		q = convert_string(ct, s, l);
 		if (q) add_to_str(&lbl, &lbl_l, q), mem_free(q);
 	}
 	if (html + 2 <= eof && (html[1] == '!' || html[1] == '?')) {
@@ -1283,22 +1495,25 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	if (t_namelen == 7 && !casecmp(t_name, "/SELECT", 7)) {
 		if (lbl) {
 			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
-			new_menu_item(lbl, order - 1), lbl = NULL;
+			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
+			else mem_free(lbl), lbl = NULL;
 		}
 		goto end_parse;
 	}
 	if (t_namelen == 7 && !casecmp(t_name, "/OPTION", 7)) {
 		if (lbl) {
 			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
-			new_menu_item(lbl, order - 1), lbl = NULL;
+			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
+			else mem_free(lbl), lbl = NULL;
 		}
 		goto see;
 	}
 	if (t_namelen == 6 && !casecmp(t_name, "OPTION", 6)) {
-		unsigned char *v;
+		unsigned char *v, *vx;
 		if (lbl) {
 			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
-			new_menu_item(lbl, order - 1), lbl = NULL;
+			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
+			else mem_free(lbl), lbl = NULL;
 		}
 		if (has_attr(t_attr, "disabled")) goto see;
 		if (preselect == -1 && has_attr(t_attr, "selected")) preselect = order;
@@ -1309,22 +1524,25 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 			val = vv;
 		}
 		val[order++] = v;
-		if ((v = get_attr_val(t_attr, "label"))) new_menu_item(v, order - 1);
-		else lbl = init_str(), lbl_l = 0;
+		if ((vx = get_attr_val(t_attr, "label"))) new_menu_item(vx, order - 1, 0);
+		if (!v || !vx) {
+			lbl = init_str(), lbl_l = 0;
+			nnmi = !!vx;
+		}
 		goto see;
 	}
 	if ((t_namelen == 8 && !casecmp(t_name, "OPTGROUP", 8)) || (t_namelen == 9 && !casecmp(t_name, "/OPTGROUP", 9))) {
-		char *la;
 		if (lbl) {
 			if (!val[order - 1]) val[order - 1] = stracpy(lbl);
-			new_menu_item(lbl, order - 1), lbl = NULL;
+			if (!nnmi) new_menu_item(lbl, order - 1, 1), lbl = NULL;
+			else mem_free(lbl), lbl = NULL;
 		}
-		if (group) new_menu_item(NULL, -1), group = 0;
+		if (group) new_menu_item(NULL, -1, 0), group = 0;
 	}
 	if (t_namelen == 8 && !casecmp(t_name, "OPTGROUP", 8)) {
 		char *la;
 		if (!(la = get_attr_val(t_attr, "label"))) if (!(la = stracpy(""))) goto see;
-		new_menu_item(la, -1);
+		new_menu_item(la, -1, 0);
 		group = 1;
 	}
 	goto see;
@@ -1339,9 +1557,8 @@ int do_html_select(unsigned char *attr, unsigned char *html, unsigned char *eof,
 		goto abort;
 	}
 	memset(lbls, 0, order * sizeof(char *));
-	fc->form_num = form_num;
-	fc->ctrl_num = form.ctrl_num++;
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->form_num = last_form_tag - startf;
+	fc->ctrl_num = attr - last_form_tag;
 	fc->position = attr - startf;
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
@@ -1375,10 +1592,11 @@ void html_textarea(unsigned char *a)
 void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *eof, unsigned char **end, void *f)
 {
 	struct form_control *fc;
-	unsigned char *p, *en, *t_name;
+	unsigned char *p, *t_name;
 	int t_namelen;
 	int cols, rows;
 	int i;
+	find_form_for_input(attr);
 	while (html < eof && (*html == '\n' || *html == '\r')) html++;
 	p = html;
 	while (p < eof && *p != '<') {
@@ -1393,9 +1611,8 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 	if (t_namelen != 9 || casecmp(t_name, "/TEXTAREA", 9)) goto pp;
 	if (!(fc = mem_alloc(sizeof(struct form_control)))) return;
 	memset(fc, 0, sizeof(struct form_control));
-	fc->form_num = form_num;
-	fc->ctrl_num = form.ctrl_num++;
-	fc->g_ctrl_num = g_ctrl_num++;
+	fc->form_num = last_form_tag - startf;
+	fc->ctrl_num = attr - last_form_tag;
 	fc->position = attr - startf;
 	fc->method = form.method;
 	fc->action = stracpy(form.action);
@@ -1431,30 +1648,213 @@ void do_html_textarea(unsigned char *attr, unsigned char *html, unsigned char *e
 void html_iframe(unsigned char *a)
 {
 	unsigned char *name, *url;
-	if (!(url = get_attr_val(a, "src"))) return;
+	if (!(url = get_url_val(a, "src"))) return;
 	if (!(name = get_attr_val(a, "name"))) name = stracpy("");
-	if (*name) put_link_line("IFrame: ", name, url);
-	else put_link_line("", "IFrame", url);
+	if (*name) put_link_line("IFrame: ", name, url, d_opt->framename);
+	else put_link_line("", "IFrame", url, d_opt->framename);
 	mem_free(name);
 	mem_free(url);
 }
 
+void html_noframes(unsigned char *a)
+{
+	if (d_opt->frames) html_skip(a);
+}
+
 void html_frame(unsigned char *a)
 {
-	unsigned char *name, *url;
-	if (!(url = get_attr_val(a, "src"))) return;
-	if (!(name = get_attr_val(a, "name"))) name = stracpy(url);
-	put_link_line("Frame: ", name, url);
+	unsigned char *name, *u2, *url;
+	if (!(u2 = get_url_val(a, "src"))) {
+		url = stracpy("");
+	} else {
+		url = join_urls(format.href_base, u2);
+		mem_free(u2);
+	}
+	if (!url) return;
+	name = get_attr_val (a, "name");
+	if (!name)
+		name = stracpy(url);
+	else if (!name[0]) { /* When name doesn't have a value */
+		mem_free(name);
+		name = stracpy(url);
+	}
+	if (!d_opt->frames || !html_top.frameset) put_link_line("Frame: ", name, url, "");
+	else {
+		struct frame_param fp;
+		fp.name = name;
+		fp.url = url;
+		fp.parent = html_top.frameset;
+		if (special_f(ff, SP_USED, NULL)) special_f(ff, SP_FRAME, &fp);
+	}
 	mem_free(name);
 	mem_free(url);
 }
 
+void parse_frame_widths(unsigned char *a, int ww, int www, int **op, int *olp)
+{
+	unsigned char *aa;
+	int q, qq, i, d, nn;
+	unsigned long n;
+	int *oo, *o;
+	int ol;
+	ol = 0;
+	o = DUMMY;
+	new_ch:
+	while (WHITECHAR(*a)) a++;
+	n = strtoul(a, (char **)&a, 10);
+	q = n;
+	if (*a == '%') q = q * ww / 100;
+	else if (*a != '*') q = (q + (www - 1) / 2) / www;
+	else if (!(q = -q)) q = -1;
+	if ((oo = mem_realloc(o, (ol + 1) * sizeof(int)))) (o = oo)[ol++] = q;
+	if ((aa = strchr(a, ','))) {
+		a = aa + 1;
+		goto new_ch;
+	}
+	*op = o;
+	*olp = ol;
+	q = 2 * ol - 1;
+	for (i = 0; i < ol; i++) if (o[i] > 0) q += o[i] - 1;
+	if (q >= ww) {
+		distribute:
+		for (i = 0; i < ol; i++) if (o[i] < 1) o[i] = 1;
+		q -= ww;
+		d = 0;
+		for (i = 0; i < ol; i++) d += o[i];
+		qq = q;
+		for (i = 0; i < ol; i++) {
+			q -= o[i] - o[i] * (d - qq) / d;
+			do_not_optimize_here(&d);
+				/* SIGH! gcc 2.7.2.* has an optimizer bug! */
+			o[i] = o[i] * (d - qq) / d;
+		}
+		while (q) {
+			nn = 0;
+			for (i = 0; i < ol; i++) {
+				if (q < 0) o[i]++, q++, nn = 1;
+				if (q > 0 && o[i] > 1) o[i]--, q--, nn = 1;
+				if (!q) break;
+			}
+			if (!nn) break;
+		}
+	} else {
+		int nn = 0;
+		for (i = 0; i < ol; i++) if (o[i] < 0) nn = 1;
+		if (!nn) goto distribute;
+		if (!(oo = mem_alloc(ol * sizeof(int)))) {
+			*olp = 0;
+			return;
+		}
+		memcpy(oo, o, ol * sizeof(int));
+		for (i = 0; i < ol; i++) if (o[i] < 1) o[i] = 1;
+		q = ww - q;
+		d = 0;
+		for (i = 0; i < ol; i++) if (oo[i] < 0) d += -oo[i];
+		qq = q;
+		for (i = 0; i < ol; i++) if (oo[i] < 0) {
+			o[i] += (-oo[i] * qq / d);
+			q -= (-oo[i] * qq / d);
+		}
+		if (q < 0) internal("parse_frame_widths: q < 0");
+		for (i = 0; i < ol; i++) if (oo[i] < 0) {
+			if (q) o[i]++, q--;
+		}
+		if (q > 0) internal("parse_frame_widths: q > 0");
+		mem_free(oo);
+	}
+	for (i = 0; i < ol; i++) if (!o[i]) {
+		int j;
+		int m = 0;
+		int mj = 0;
+		for (j = 0; j < ol; j++) if (o[j] > m) m = o[j], mj = j;
+		if (m) o[i] = 1, o[mj]--;
+	}
+}
+
+void html_frameset(unsigned char *a)
+{
+	int x, y;
+	struct frameset_param fp;
+	unsigned char *c, *d;
+	if (!d_opt->frames || !special_f(ff, SP_USED, NULL)) return;
+	if (!(c = get_attr_val(a, "cols"))) if (!(c = stracpy("100%"))) return;
+	if (!(d = get_attr_val(a, "rows"))) if (!(d = stracpy("100%"))) {
+		mem_free(c);
+		return;
+	}
+	if (!html_top.frameset) {
+		x = d_opt->xw;
+		y = d_opt->yw;
+	} else {
+		struct frameset_desc *f = html_top.frameset;
+		if (f->yp >= f->y) goto free_cd;
+		x = f->f[f->xp + f->yp * f->x].xw;
+		y = f->f[f->xp + f->yp * f->x].yw;
+	}
+	parse_frame_widths(c, x, HTML_FRAME_CHAR_WIDTH, &fp.xw, &fp.x);
+	parse_frame_widths(d, y, HTML_FRAME_CHAR_HEIGHT, &fp.yw, &fp.y);
+	fp.parent = html_top.frameset;
+	if (fp.x && fp.y) html_top.frameset = special_f(ff, SP_FRAMESET, &fp);
+	mem_free(fp.xw);
+	mem_free(fp.yw);
+	free_cd:
+	mem_free(c);
+	mem_free(d);
+}
+
+/*void html_frameset(unsigned char *a)
+{
+	int w;
+	int horiz = 0;
+	struct frameset_param *fp;
+	unsigned char *c, *d;
+	if (!d_opt->frames || !special_f(ff, SP_USED, NULL)) return;
+	if (!(c = get_attr_val(a, "cols"))) {
+		horiz = 1;
+		if (!(c = get_attr_val(a, "rows"))) return;
+	}
+	if (!(fp = mem_alloc(sizeof(struct frameset_param)))) goto f;
+	fp->n = 0;
+	fp->horiz = horiz;
+	par_format.leftmargin = par_format.rightmargin = 0;
+	d = c;
+	while (1) {
+		while (WHITECHAR(*d)) d++;
+		if (!*d) break;
+		if (*d == ',') {
+			d++;
+			continue;
+		}
+		if ((w = parse_width(d, 1)) != -1) {
+			struct frameset_param *fpp;
+			if ((fpp = mem_realloc(fp, sizeof(struct frameset_param) + (fp->n + 1) * sizeof(int)))) {
+				fp = fpp;
+				fp->width[fp->n++] = w;
+			}
+		}
+		if (!(d = strchr(d, ','))) break;
+		d++;
+	}
+	fp->parent = html_top.frameset;
+	if (fp->n) html_top.frameset = special_f(ff, SP_FRAMESET, fp);
+	mem_free(fp);
+	f:
+	mem_free(c);
+}*/
+
 void html_link(unsigned char *a)
 {
 	unsigned char *name, *url;
-	if (!(url = get_attr_val(a, "href"))) return;
+	if ((name = get_attr_val(a, "type"))) {
+		if (casecmp(a, "text/css", 8)) {
+			mem_free(name);
+			return;
+		}
+		mem_free(name);
+	}
+	if (!(url = get_url_val(a, "href"))) return;
 	if (!(name = get_attr_val(a, "rel"))) if (!(name = get_attr_val(a, "rev"))) name = stracpy(url);
-	put_link_line("Link: ", name, url);
+	if (strcasecmp(name, "STYLESHEET") && strcasecmp(name, "made") && strcasecmp(name, "SHORTCUT ICON")) put_link_line("Link: ", name, url, format.target_base);
 	mem_free(name);
 	mem_free(url);
 }
@@ -1524,15 +1924,18 @@ struct element_info elements[] = {
 	{"TD",		html_td,	0, 0},
 	{"TH",		html_th,	0, 0},
 
-	{"FORM",	html_form,	1, 1},
+	{"FORM",	html_form,	1, 0},
 	{"INPUT",	html_input,	0, 1},
 	{"TEXTAREA",	html_textarea,	0, 1},
 	{"SELECT",	html_select,	0, 0},
 	{"OPTION",	html_option,	1, 1},
 	{"BUTTON",	html_button,	0, 0},
 
+	{"LINK",	html_link,	1, 1},
 	{"IFRAME",	html_iframe,	1, 1},
 	{"FRAME",	html_frame,	1, 1},
+	{"FRAMESET",	html_frameset,	1, 0},
+	{"NOFRAMES",	html_noframes,	0, 0},
 
 	{NULL,		NULL, 0, 0},
 };
@@ -1559,9 +1962,9 @@ unsigned char *skip_comment(unsigned char *html, unsigned char *eof)
 void process_head(unsigned char *head)
 {
 	unsigned char *r, *p;
-	if ((r = parse_http_header(head, "Refresh"))) {
+	if ((r = parse_http_header(head, "Refresh", NULL))) {
 		if ((p = parse_header_param(r, "URL"))) {
-			put_link_line("Refresh: ", p, p);
+			put_link_line("Refresh: ", p, p, d_opt->framename);
 			mem_free(p);
 		}
 		mem_free(r);
@@ -1614,8 +2017,10 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 			}
 			html++;
 			if (!(pos + (html-lt-1))) goto skip_w; /* ??? */
-			if (*(html - 1) == ' ') put_chrs(lt, html - lt, put_chars, f);
-			else {
+			if (*(html - 1) == ' ') {
+				if (html < eof && !WHITECHAR(*html)) continue;	/* BIG performance win; not sure if it doesn't cause any bug */
+				put_chrs(lt, html - lt, put_chars, f);
+			} else {
 				put_chrs(lt, html - 1 - lt, put_chars, f);
 				put_chrs(" ", 1, put_chars, f);
 			}
@@ -1776,7 +2181,7 @@ void parse_html(unsigned char *html, unsigned char *eof, void (*put_chars)(void
 	was_br = 0;
 }
 
-int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, unsigned char *tag, struct menu_item **menu, struct memory_list **ml, unsigned char *href_base, unsigned char *target_base, int to, int def)
+int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, unsigned char *tag, struct menu_item **menu, struct memory_list **ml, unsigned char *href_base, unsigned char *target_base, int to, int def, int hdef)
 {
 	unsigned char *name, *attr, *al, *label, *href, *target;
 	int namelen, lblen;
@@ -1789,7 +2194,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 	struct conv_table *ct;
 	if (head) add_to_str(&hd, &hdl, head);
 	scan_http_equiv(s, eof, &hd, &hdl, NULL);
-	ct = get_convert_table(hd, to, def, NULL, NULL);
+	ct = get_convert_table(hd, to, def, NULL, NULL, hdef);
 	mem_free(hd);
 	if (!(*menu = mem_alloc(sizeof(struct menu_item)))) return -1;
 	memset(*menu, 0, sizeof(struct menu_item));
@@ -1868,11 +2273,11 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		else label = NULL;
 	} else if (namelen == 4 && !casecmp(name, "/MAP", 4)) goto done;
 	else goto se2;
-	if (!(href = get_attr_val(attr, "href"))) {
+	if (!(href = get_url_val(attr, "href"))) {
 		if (label) mem_free(label);
 		goto se2;
 	}
-	if (!(target = get_attr_val(attr, "target")) && !(target = stracpy(target_base)))
+	if (!(target = get_target(attr)) && !(target = stracpy(target_base)))
 		target = stracpy("");
 	if (!(ld = mem_alloc(sizeof(struct link_def)))) {
 		if (label) mem_free(label);
@@ -1910,6 +2315,7 @@ int get_image_map(unsigned char *head, unsigned char *s, unsigned char *eof, uns
 		memset(&nm[nmenu], 0, 2 * sizeof(struct menu_item));
 		nm[nmenu].text = label;
 		nm[nmenu].rtext = "";
+		nm[nmenu].hotkey = "";
 		nm[nmenu].func = MENU_FUNC map_selected;
 		nm[nmenu].data = ld;
 		nm[++nmenu].text = NULL;
@@ -1956,8 +2362,13 @@ void scan_http_equiv(unsigned char *s, unsigned char *eof, unsigned char **head,
 	if (namelen != 4 && !casecmp(name, "META", 4)) goto se;
 	if (!(he = get_attr_val(attr, "http-equiv"))) goto se;
 	c = get_attr_val(attr, "content");
-	add_to_str(head, hdl, he), mem_free(he);
+	add_to_str(head, hdl, he);
 	if (c) add_to_str(head, hdl, ": "), add_to_str(head, hdl, c), mem_free(c);
+	if (!strcasecmp(he, "content-type")) {
+		c = get_attr_val(attr, "charset");
+		if (c) add_to_str(head, hdl, ";charset="), add_to_str(head, hdl, c), mem_free(c);
+	}
+	mem_free(he);
 	add_to_str(head, hdl, "\r\n");
 	goto se;
 }
