@@ -9,7 +9,7 @@ static inline int color_distance(struct rgb *c1, struct rgb *c2)
 }
 
 struct rgb palette[] = {
-	{0x00, 0x00, 0x00},
+/*	{0x00, 0x00, 0x00},
 	{0x80, 0x00, 0x00},
 	{0x00, 0x80, 0x00},
 	{0x80, 0x80, 0x00},
@@ -24,11 +24,43 @@ struct rgb palette[] = {
 	{0x00, 0x00, 0xff},
 	{0xff, 0x00, 0xff},
 	{0x00, 0xff, 0xff},
+	{0xff, 0xff, 0xff},*/
+	/*{0x00, 0x00, 0x00},
+	{0xaa, 0x00, 0x00},
+	{0x00, 0xaa, 0x00},
+	{0xaa, 0x55, 0x00},
+	{0x00, 0x00, 0xaa},
+	{0xaa, 0x00, 0xaa},
+	{0x00, 0xaa, 0xaa},
+	{0xaa, 0xaa, 0xaa},
+	{0x55, 0x55, 0x55},
+	{0xff, 0x55, 0x55},
+	{0x55, 0xff, 0x55},
+	{0xff, 0xff, 0x55},
+	{0x55, 0x55, 0xff},
+	{0xff, 0x55, 0xff},
+	{0x55, 0xff, 0xff},
+	{0xff, 0xff, 0xff},*/
+	{0x00, 0x00, 0x00},
+	{0x80, 0x00, 0x00},
+	{0x00, 0x80, 0x00},
+	{0xaa, 0x55, 0x00},
+	{0x00, 0x00, 0x80},
+	{0x80, 0x00, 0x80},
+	{0x00, 0x80, 0x80},
+	{0xaa, 0xaa, 0xaa},
+	{0x55, 0x55, 0x55},
+	{0xff, 0x55, 0x55},
+	{0x55, 0xff, 0x55},
+	{0xff, 0xff, 0x55},
+	{0x55, 0x55, 0xff},
+	{0xff, 0x55, 0xff},
+	{0x55, 0xff, 0xff},
 	{0xff, 0xff, 0xff},
 	{-1, -1, -1}
 };
 
-struct rgb rgbcache = {0, 0, 0};
+/*struct rgb rgbcache = {0, 0, 0};
 int rgbcache_c = 0;
 
 static inline int find_nearest_color(struct rgb *r, int l)
@@ -40,6 +72,43 @@ static inline int find_nearest_color(struct rgb *r, int l)
 	for (i = 0; i < l; i++) if ((dst = color_distance(r, &palette[i])) < dist)
 		dist = dst, min = i;
 	return min;
+}*/
+
+struct rgb_cache_entry {
+	int color;
+	int l;
+	struct rgb rgb;
+};
+
+#define RGB_HASH_SIZE 4096
+
+#define HASH_RGB(r, l) ((((r)->r << 3) + ((r)->g << 2) + (r)->b + (l)) & (RGB_HASH_SIZE - 1))
+
+static inline int find_nearest_color(struct rgb *r, int l)
+{
+	int dist, dst, min, i;
+	static struct rgb_cache_entry rgb_cache[RGB_HASH_SIZE];
+	static int cache_init = 0;
+	int h;
+	if (!cache_init) goto initialize;
+	back:
+	h = HASH_RGB(r, l);
+	if (rgb_cache[h].color != -1 && rgb_cache[h].l == l && rgb_cache[h].rgb.r == r->r && rgb_cache[h].rgb.g == r->g && rgb_cache[h].rgb.b == r->b) return rgb_cache[h].color;
+	dist = 0xffffff;
+	min = 0;
+	for (i = 0; i < l; i++) if ((dst = color_distance(r, &palette[i])) < dist)
+		dist = dst, min = i;
+	rgb_cache[h].color = min;
+	rgb_cache[h].l = l;
+	rgb_cache[h].rgb.r = r->r;
+	rgb_cache[h].rgb.g = r->g;
+	rgb_cache[h].rgb.b = r->b;
+	return min;
+
+	initialize:
+	for (h = 0; h < RGB_HASH_SIZE; h++) rgb_cache[h].color = -1;
+	cache_init = 1;
+	goto back;
 }
 
 static inline int fg_color(int fg, int bg)
@@ -161,7 +230,7 @@ int xxpand_line(struct part *p, int y, int x)
 {
 	return xpand_line(p, y, x);
 }
-				
+
 static inline void set_hline(struct part *p, int x, int y, int xl, unsigned char *d, unsigned c, int spc)
 {
 	if (xpand_lines(p, y)) return;
@@ -269,7 +338,6 @@ int split_line(struct part *p)
 		if ((POS(i, p->cy) & 0xff) == ' ') goto split;
 	for (i = p->cx - 1; i > rm(par_format) && i > par_format.leftmargin; i--)
 		if ((POS(i, p->cy) & 0xff) == ' ') goto split;*/
-	r:
 	if (p->cx + par_format.rightmargin > p->x) p->x = p->cx + par_format.rightmargin;
 	/*if (p->y < p->cy + 1) p->y = p->cy + 1;
 	p->cy++; p->cx = -1;
@@ -405,6 +473,10 @@ void put_chars_conv(struct part *p, unsigned char *c, int l)
 
 void put_chars(struct part *p, unsigned char *c, int l)
 {
+	static struct text_attrib_beginning ta_cache = { -1, { 0, 0, 0 }, { 0, 0, 0 } };
+	static int bg_cache;
+	static int fg_cache;
+
 	int bg, fg;
 	int i;
 	struct link *link;
@@ -416,6 +488,32 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		last_tag_for_newline = (void *)&p->data->tags;
 	}
 	if (p->cx == -1) p->cx = par_format.leftmargin;
+	if (last_link || last_image || last_form || format.link || format.image || format.form) goto process_link;
+	no_l:
+	/*printf("%d %d\n",p->cx, p->cy);*/
+	if (memcmp(&ta_cache, &format, sizeof(struct text_attrib_beginning))) goto format_change;
+	bg = bg_cache, fg = fg_cache;
+	end_format_change:
+	if (p->cx == par_format.leftmargin && *c == ' ' && par_format.align != AL_NO) c++, l--;
+	if (p->y < p->cy + 1) p->y = p->cy + 1;
+	set_hline(p, p->cx, p->cy, l, c, (((fg&0x08)<<3)|(bg<<3)|(fg&0x07))<<8, 1);
+	p->cx += l;
+	nobreak = 0;
+	if (par_format.align != AL_NO)
+		while (p->cx > rm(par_format) && p->cx > par_format.leftmargin) {
+			int x;
+			/*if (p->cx > p->x) {
+				p->x = p->cx + par_format.rightmargin;
+				if (c[l - 1] == ' ') p->x--;
+			}*/
+			if (!(x = split_line(p))) break;
+			/*if (LEN(p->cy-1) > p->x) p->x = LEN(p->cy-1);*/
+			align_line(p, p->cy - 1);
+			nobreak = x - 1;
+		}
+	if ((p->xa += l) - (c[l-1] == ' ' && par_format.align != AL_NO) + par_format.leftmargin + par_format.rightmargin > p->xmax) p->xmax = p->xa - (c[l-1] == ' ' && par_format.align != AL_NO) + par_format.leftmargin + par_format.rightmargin;
+	return;
+	process_link:
 	if ((last_link /*|| last_target*/ || last_image || last_form) &&
 	    !xstrcmp(format.link, last_link) && !xstrcmp(format.target, last_target) &&
 	    !xstrcmp(format.image, last_image) && format.form == last_form) {
@@ -428,17 +526,35 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		goto set_link;
 		x:;
 	} else {
-		if (!last_link && !last_image && !last_form && !format.link && !format.image && !format.form) goto no_l;
-		if (!p->data) goto no_l;
 		if (last_link) mem_free(last_link);	/* !!! FIXME: optimize */
 		if (last_target) mem_free(last_target);
 		if (last_image) mem_free(last_image);
+		last_link = last_target = last_image = NULL;
+		last_form = NULL;
+		if (!(format.link || format.image || format.form)) goto no_l;
+		if (d_opt->num_links) {
+			unsigned char s[64];
+			unsigned char *fl = format.link, *ft = format.target, *fi = format.image;
+			struct form_control *ff = format.form;
+			format.link = format.target = format.image = NULL;
+			format.form = NULL;
+			s[0] = '[';
+			snzprint(s + 1, 62, p->link_num);
+			strcat(s, "]");
+			put_chars(p, s, strlen(s));
+			if (ff && ff->type == FC_TEXTAREA) line_break(p);
+			if (p->cx == -1) p->cx = par_format.leftmargin;
+			format.link = fl, format.target = ft, format.image = fi;
+			format.form = ff;
+		}
+		p->link_num++;
 		last_link = stracpy(format.link);
 		last_target = stracpy(format.target);
 		last_image = stracpy(format.image);
 		last_form = format.form;
-		if (!(last_link || last_image || last_form)) goto no_l;
+		if (!p->data) goto no_l;
 		if (!(link = new_link(p->data))) goto no_l;
+		link->num = p->link_num - 1;
 		link->pos = DUMMY;
 		if (!last_form) {
 			link->type = L_LINK;
@@ -447,6 +563,7 @@ void put_chars(struct part *p, unsigned char *c, int l)
 		} else {
 			link->type = last_form->type == FC_TEXT || last_form->type == FC_PASSWORD || last_form->type == FC_FILE ? L_FIELD : last_form->type == FC_TEXTAREA ? L_AREA : last_form->type == FC_CHECKBOX || last_form->type == FC_RADIO ? L_CHECKBOX : last_form->type == FC_SELECT ? L_SELECT : L_BUTTON;
 			link->form = last_form;
+			link->target = stracpy(last_form->target);
 		}
 		link->where_img = stracpy(last_image);
 		if (link->type != L_FIELD && link->type != L_AREA) {
@@ -468,34 +585,20 @@ void put_chars(struct part *p, unsigned char *c, int l)
 			link->n += i;
 		}
 	}
-	no_l:
-	/*printf("%d %d\n",p->cx, p->cy);*/
-	bg = find_nearest_color(&format.bg, 8);		/* !!! FIXME: use cache! */
-	fg = find_nearest_color(&format.fg, 16);
-	fg = fg_color(fg, bg);
-	if (format.attr & AT_ITALIC) fg = fg ^ 0x01;
-	if (format.attr & AT_UNDERLINE) fg = (fg ^ 0x04) | 0x08;
-	if (format.attr & AT_BOLD) fg = fg | 0x08;
-	fg = fg_color(fg, bg);
-	if (format.attr & AT_GRAPHICS) bg = bg | 0x10;
-	if (p->cx == par_format.leftmargin && *c == ' ' && par_format.align != AL_NO) c++, l--;
-	if (p->y < p->cy + 1) p->y = p->cy + 1;
-	set_hline(p, p->cx, p->cy, l, c, (((fg&0x08)<<3)|(bg<<3)|(fg&0x07))<<8, 1);
-	p->cx += l;
-	nobreak = 0;
-	if (par_format.align != AL_NO)
-		while (p->cx > rm(par_format) && p->cx > par_format.leftmargin) {
-			int x;
-			/*if (p->cx > p->x) {
-				p->x = p->cx + par_format.rightmargin;
-				if (c[l - 1] == ' ') p->x--;
-			}*/
-			if (!(x = split_line(p))) break;
-			/*if (LEN(p->cy-1) > p->x) p->x = LEN(p->cy-1);*/
-			align_line(p, p->cy - 1);
-			nobreak = x - 1;
-		}
-	if ((p->xa += l) - (c[l-1] == ' ' && par_format.align != AL_NO) + par_format.leftmargin + par_format.rightmargin > p->xmax) p->xmax = p->xa - (c[l-1] == ' ' && par_format.align != AL_NO) + par_format.leftmargin + par_format.rightmargin;
+	goto no_l;
+
+		format_change:
+		bg = find_nearest_color(&format.bg, 8);
+		fg = find_nearest_color(&format.fg, 16);
+		fg = fg_color(fg, bg);
+		if (format.attr & AT_ITALIC) fg = fg ^ 0x01;
+		if (format.attr & AT_UNDERLINE) fg = (fg ^ 0x04) | 0x08;
+		if (format.attr & AT_BOLD) fg = fg | 0x08;
+		fg = fg_color(fg, bg);
+		if (format.attr & AT_GRAPHICS) bg = bg | 0x10;
+		memcpy(&ta_cache, &format, sizeof(struct text_attrib_beginning));
+		fg_cache = fg; bg_cache = bg;
+		goto end_format_change;
 }
 
 void line_break(struct part *p)
@@ -530,13 +633,17 @@ void html_init(struct part *p)
 	/* !!! FIXME: background */
 }
 
+int g_ctrl_num;
+
 void html_form_control(struct part *p, struct form_control *fc)
 {
 	if (!p->data) {
-		destroy_fc(fc);
-		mem_free(fc);
+		/*destroy_fc(fc);
+		mem_free(fc);*/
+		add_to_list(p->uf, fc);
 		return;
 	}
+	fc->g_ctrl_num = g_ctrl_num++;
 	if (fc->type == FC_TEXT || fc->type == FC_PASSWORD || fc->type == FC_TEXTAREA) {
 		unsigned char *dv = convert_string(convert_table, fc->default_value, strlen(fc->default_value));
 		if (dv) {
@@ -547,11 +654,50 @@ void html_form_control(struct part *p, struct form_control *fc)
 	add_to_list(p->data->forms, fc);
 }
 
+void add_frameset_entry(struct frameset_desc *fsd, struct frameset_desc *subframe, unsigned char *name, unsigned char *url)
+{
+	if (fsd->yp >= fsd->y) return;
+	fsd->f[fsd->xp + fsd->yp * fsd->x].subframe = subframe;
+	fsd->f[fsd->xp + fsd->yp * fsd->x].name = stracpy(name);
+	fsd->f[fsd->xp + fsd->yp * fsd->x].url = stracpy(url);
+	if (++fsd->xp >= fsd->x) fsd->xp = 0, fsd->yp++;
+}
+
+struct frameset_desc *create_frameset(struct f_data *fda, struct frameset_param *fp)
+{
+	int i;
+	struct frameset_desc *fd;
+	if (!fp->x || !fp->y) {
+		internal("zero size of frameset");
+		return NULL;
+	}
+	if (!(fd = mem_alloc(sizeof(struct frameset_desc) + fp->x * fp->y * sizeof(struct frame_desc)))) return NULL;
+	memset(fd, 0, sizeof(struct frameset_desc) + fp->x * fp->y * sizeof(struct frame_desc));
+	fd->n = fp->x * fp->y;
+	fd->x = fp->x;
+	fd->y = fp->y;
+	for (i = 0; i < fd->n; i++) {
+		fd->f[i].xw = fp->xw[i % fp->x];
+		fd->f[i].yw = fp->yw[i / fp->x];
+	}
+	if (fp->parent) add_frameset_entry(fp->parent, fd, NULL, NULL);
+	else if (!fda->frame_desc) fda->frame_desc = fd;
+	     else mem_free(fd), fd = NULL;
+	return fd;
+}
+
+void create_frame(struct frame_param *fp)
+{
+	add_frameset_entry(fp->parent, NULL, fp->name, fp->url);
+}
+
 void *html_special(struct part *p, int c, ...)
 {
 	va_list l;
 	unsigned char *t;
 	struct form_control *fc;
+	struct frameset_param *fsp;
+	struct frame_param *fp;
 	va_start(l, c);
 	switch (c) {
 		case SP_TAG:
@@ -566,6 +712,13 @@ void *html_special(struct part *p, int c, ...)
 			return convert_table;
 		case SP_USED:
 			return (void *)!!p->data;
+		case SP_FRAMESET:
+			fsp = va_arg(l, struct frameset_param *);
+			return create_frameset(p->data, fsp);
+		case SP_FRAME:
+			fp = va_arg(l, struct frame_param *);
+			create_frame(fp);
+			break;
 		default:
 			internal("html_special: unknown code %d", c);
 	}
@@ -580,7 +733,7 @@ void do_format(char *start, char *end, struct part *part, unsigned char *head)
 
 int margin;
 
-struct part *format_html_part(unsigned char *start, unsigned char *end, int align, int m, int width, struct f_data *data, int xs, int ys, unsigned char *head)
+struct part *format_html_part(unsigned char *start, unsigned char *end, int align, int m, int width, struct f_data *data, int xs, int ys, unsigned char *head, int link_num)
 {
 	struct part *p;
 	struct html_element *e;
@@ -589,16 +742,7 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	/*struct tag *ltn = last_tag_for_newline;*/
 	int lm = margin;
 	int ef = empty_format;
-	int last_form_num;
-	int onx, onxw;
-	struct form last_form;
-	int last_g_ctrl_num;
-	if (!data) {
-		last_form_num = form_num;
-		memcpy(&last_form, &form, sizeof(struct form));
-		last_form.action = stracpy(form.action);
-		last_g_ctrl_num = g_ctrl_num;
-	}
+	struct form_control *fc;
 	if (data) {
 		struct node *n;
 		if ((n = mem_alloc(sizeof(struct node)))) {
@@ -618,6 +762,7 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	if (last_image) mem_free(last_image);
 	if (last_target) mem_free(last_target);
 	last_link = last_image = last_target = NULL;
+	last_form = NULL;
 	nobreak = 1;
 	if (!(p = mem_alloc(sizeof(struct part)))) goto ret;
 	p->x = p->y = 0;
@@ -627,6 +772,8 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 	p->bgcolor = find_nearest_color(&par_format.bgcolor, 8);
 	p->spaces = DUMMY;
 	p->spl = 0;
+	p->link_num = link_num;
+	init_list(p->uf);
 	html_stack_dup();
 	e = &html_top;
 	html_top.dontkill = 2;
@@ -660,19 +807,16 @@ struct part *format_html_part(unsigned char *start, unsigned char *end, int alig
 		struct node *n = data->nodes.next;
 		n->yw = ys - n->y + p->y;
 	}
+	foreach(fc, p->uf) destroy_fc(fc);
+	free_list(p->uf);
 	ret:
-	if (!data) {
-		form_num = last_form_num;
-		if (form.action) mem_free(form.action);
-		memcpy(&form, &last_form, sizeof(struct form));
-		g_ctrl_num = last_g_ctrl_num;
-	}
 	last_link_to_move = llm;
 	last_tag_to_move = ltm;
 	/*last_tag_for_newline = ltn;*/
 	margin = lm;
 	empty_format = ef;
 	last_link = last_image = last_target = NULL;
+	last_form = NULL;
 	return p;
 }
 
@@ -710,23 +854,26 @@ void push_base_format(unsigned char *url, struct document_options *opt)
 	html_top.dontkill = 1;
 }
 
-struct conv_table *get_convert_table(unsigned char *head, int to, int def, int *frm, int *aa)
+struct conv_table *get_convert_table(unsigned char *head, int to, int def, int *frm, int *aa, int hard)
 {
 	int from = -1;
 	unsigned char *a, *b;
-	if ((a = parse_http_header(head, "Content-Charset"))) {
+	if (head && (a = parse_http_header(head, "Content-Charset", NULL))) {
 		from = get_cp_index(a);
 		mem_free(a);
 	}
-	if (from == -1) if ((a = parse_http_header(head, "Content-Type"))) {
+	if (from == -1 && head && (a = parse_http_header(head, "Content-Type", NULL))) {
 		if ((b = parse_header_param(a, "charset"))) {
 			from = get_cp_index(b);
 			mem_free(b);
 		}
 		mem_free(a);
 	}
-	if (aa) *aa = from == -1;
-	if (from == -1) from = def;
+	if (aa) {
+		*aa = from == -1;
+		if (hard && !*aa) *aa = 2;
+	}
+	if (hard || from == -1) from = def;
 	if (frm) *frm = from;
 	return get_translation_table(from, to);
 }
@@ -744,6 +891,7 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 	int i;
 	d_opt = &screen->opt;
 	screen->use_tag = ce->count;
+	defrag_entry(ce);
 	fr = ce->frag.next;
 	if ((void *)fr == &ce->frag || fr->offset || !fr->length) start = NULL, end = NULL;
 	else start = fr->data, end = fr->data + fr->length;
@@ -752,15 +900,18 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 	head = init_str(), hdl = 0;
 	if (ce->head) add_to_str(&head, &hdl, ce->head);
 	scan_http_equiv(start, end, &head, &hdl, &t);
-	convert_table = get_convert_table(head, screen->opt.cp, screen->opt.assume_cp, &screen->cp, &screen->ass);
+	convert_table = get_convert_table(head, screen->opt.cp, screen->opt.assume_cp, &screen->cp, &screen->ass, screen->opt.hard_assume);
+	i = d_opt->plain; d_opt->plain = 0;
 	screen->title = convert_string(convert_table, t, strlen(t));
+	d_opt->plain = i;
 	mem_free(t);
 	push_base_format(url, &screen->opt);
 	table_level = 0;
-	form_num = 0;
 	g_ctrl_num = 0;
-	memset(&form, 0, sizeof(struct form));
-	if ((rp = format_html_part(start, end, par_format.align, par_format.leftmargin, screen->opt.xw, screen, 0, 0, head))) mem_free(rp);
+	last_form_tag = NULL;
+	last_form_attr = NULL;
+	last_input_tag = NULL;
+	if ((rp = format_html_part(start, end, par_format.align, par_format.leftmargin, screen->opt.xw, screen, 0, 0, head, 1))) mem_free(rp);
 	mem_free(head);
 	screen->x = 0;
 	for (i = screen->y - 1; i >= 0; i--) {
@@ -768,7 +919,8 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 		else break;
 	}
 	for (i = 0; i < screen->y; i++) if (screen->data[i].l > screen->x) screen->x = screen->data[i].l;
-	if (form.action) mem_free(form.action);
+	if (form.action) mem_free(form.action), form.action = NULL;
+	if (form.target) mem_free(form.target), form.target = NULL;
 	kill_html_stack_item(html_stack.next);
 	if (html_stack.next != &html_stack) {
 		internal("html stack not empty after operation");
@@ -776,6 +928,7 @@ void format_html(struct cache_entry *ce, struct f_data *screen)
 	}
 	screen->bg = 007 << 8; /* !!! FIXME */
 	sort_links(screen);
+	if (screen->frame_desc) screen->frame = 1;
 	/*{
 		FILE *f = fopen("forms", "a");
 		struct form_control *form;
@@ -798,14 +951,18 @@ static inline int compare_opt(struct document_options *o1, struct document_optio
 	    o1->col == o2->col &&
 	    o1->cp == o2->cp &&
 	    o1->assume_cp == o2->assume_cp &&
+	    o1->hard_assume == o2->hard_assume &&
 	    o1->tables == o2->tables &&
 	    o1->frames == o2->frames &&
+	    o1->images == o2->images &&
 	    o1->margin == o2->margin &&
 	    o1->plain == o2->plain &&
 	    !memcmp(&o1->default_fg, &o2->default_fg, sizeof(struct rgb)) &&
 	    !memcmp(&o1->default_bg, &o2->default_bg, sizeof(struct rgb)) &&
 	    !memcmp(&o1->default_link, &o2->default_link, sizeof(struct rgb)) &&
 	    !memcmp(&o1->default_vlink, &o2->default_vlink, sizeof(struct rgb)) &&
+	    o1->num_links == o2->num_links &&
+	    o1->table_order == o2->table_order &&
 	    !strcasecmp(o1->framename, o2->framename)) return 0;
 	return 1;
 }
@@ -860,18 +1017,27 @@ void delete_unused_format_cache_entries()
 	}
 }
 
+void format_cache_reactivate(struct f_data *ce)
+{
+	del_from_list(ce);
+	add_to_list(format_cache, ce);
+}
+
 void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct document_options *opt)
 {
+	unsigned char *n;
 	struct f_data *ce;
 	struct cache_entry *cee;
 	if (!vs) return;
+	n = screen->name; screen->name = NULL;
 	detach_formatted(screen);
+	screen->name = n;
 	screen->link_bg = NULL;
 	screen->link_bg_n = 0;
 	if (vs) vs->f = screen;
 	screen->vs = vs;
 	screen->xl = screen->yl = -1;
-	screen->f1 = screen->f2 = screen->f_data = NULL;
+	screen->f_data = NULL;
 	foreach(ce, format_cache) if (!strcmp(ce->url, vs->url) && !compare_opt(&ce->opt, opt)) {
 		cee = NULL;
 		if (find_in_cache(vs->url, &cee) || !cee || cee->count != ce->use_tag) {
@@ -883,13 +1049,11 @@ void cached_format_html(struct view_state *vs, struct f_data_c *screen, struct d
 			}
 			continue;
 		}
-		del_from_list(ce);
-		add_to_list(format_cache, ce);
+		format_cache_reactivate(ce);
 		if (!ce->refcount++) format_cache_entries--;
 		screen->f_data = ce;
 		goto sx;
 	}
-	new_e:
 	if (find_in_cache(vs->url, &cee) || !cee) {
 		internal("document to format not found");
 		return;
@@ -938,10 +1102,103 @@ long formatted_info(int type)
 	return 0;
 }
 
+void add_frame_to_list(struct session *ses, struct f_data_c *fd)
+{
+	struct f_data_c *f;
+	foreach(f, ses->scrn_frames) {
+		if (f->yp > fd->yp || (f->yp == fd->yp && f->xp > fd->xp)) {
+			add_at_pos(f->prev, fd);
+			return;
+		}
+	}
+	add_at_pos((struct f_data_c *)ses->scrn_frames.prev, fd);
+}
+
+struct f_data_c *find_fd(struct session *ses, unsigned char *name, int depth, int x, int y)
+{
+	struct f_data_c *fd;
+	foreachback(fd, ses->scrn_frames) if (!strcasecmp(fd->name, name) && !fd->used) {
+		fd->used = 1;
+		fd->depth = depth;
+		return fd;
+	}
+	if (!(fd = mem_alloc(sizeof(struct f_data_c)))) return NULL;
+	memset(fd, 0, sizeof(struct f_data_c));
+	fd->used = 1;
+	fd->name = stracpy(name);
+	fd->depth = depth;
+	fd->xp = x, fd->yp = y;
+	/*add_to_list(ses->scrn_frames, fd);*/
+	add_frame_to_list(ses, fd);
+	return fd;
+}
+
+struct f_data_c *format_frame(struct session *ses, unsigned char *name, unsigned char *url, struct document_options *o, int depth)
+{
+	struct cache_entry *ce;
+	struct view_state *vs;
+	struct f_data_c *fd;
+	struct frame *fr;
+	repeat:
+	if (!(fr = ses_find_frame(ses, name))) return NULL;
+	vs = &fr->vs;
+	if (find_in_cache(vs->url, &ce) || !ce) return NULL;
+	if (ce->redirect && fr->redirect_cnt < MAX_REDIRECTS) {
+		unsigned char *u;
+		if ((u = join_urls(vs->url, ce->redirect))) {
+			fr->redirect_cnt++;
+			ses_change_frame_url(ses, name, u);
+			mem_free(u);
+			goto repeat;
+		}
+	}
+	if (!(fd = find_fd(ses, name, depth, o->xp, o->yp))) return NULL;
+	cached_format_html(vs, fd, o);
+	return fd;
+}
+
+void format_frames(struct session *ses, struct frameset_desc *fsd, struct document_options *op, int depth)
+{
+	int i, j, n;
+	struct document_options o;
+	if (depth > HTML_MAX_FRAME_DEPTH) return;
+	memcpy(&o, op, sizeof(struct document_options));
+	if (o.margin) o.margin = 1;
+	n = 0;
+	for (j = 0; j < fsd->y; j++) {
+		o.xp = op->xp;
+		for (i = 0; i < fsd->x; i++) {
+			struct f_data_c *fdc;
+			o.xw = fsd->f[n].xw;
+			o.yw = fsd->f[n].yw;
+			o.framename = fsd->f[n].name;
+			if (fsd->f[n].subframe) format_frames(ses, fsd->f[n].subframe, &o, depth + 1);
+			else if (fsd->f[n].name) {
+				fdc = format_frame(ses, fsd->f[n].name, fsd->f[n].url, &o, depth);
+				if (fdc && fdc->f_data && fdc->f_data->frame) format_frames(ses, fdc->f_data->frame_desc, &o, depth + 1);
+			}
+			o.xp += o.xw + 1;
+			n++;
+		}
+		o.yp += o.yw + 1;
+	}
+			
+	/*for (i = 0; i < fsd->n; i++) {
+		if (!fsd->horiz) o.xw = fsd->f[i].width;
+		else o.yw = fsd->f[i].width;
+		o.framename = fsd->f[i].name;
+		if (fsd->f[i].subframe) format_frames(ses, fsd->f[i].subframe, &o);
+		else format_frame(ses, fsd->f[i].name, fsd->f[i].url, &o);
+		if (!fsd->horiz) o.xp += fsd->f[i].width + 1;
+		else o.yp += fsd->f[i].width + 1;
+	}*/
+}
+
 void html_interpret(struct session *ses)
 {
 	struct view_state *l;
 	struct document_options o;
+	struct f_data_c *fd, *cf = NULL;
 	/*debug("I");*/
 	if (!ses->screen) {
 		if (!(ses->screen = mem_alloc(sizeof(struct f_data_c)))) return;
@@ -953,12 +1210,9 @@ void html_interpret(struct session *ses)
 	o.yp = 1;
 	o.xw = ses->term->x;
 	o.yw = ses->term->y - 2;
-	o.col = get_term_spec(ses->term->term)->col;
-	o.cp = get_term_spec(ses->term->term)->charset;
-	o.assume_cp = ses->assume_cp;
-	o.tables = enable_html_tables;
-	o.frames = enable_html_frames;
-	o.margin = default_left_margin;
+	o.col = ses->term->spec->col;
+	o.cp = ses->term->spec->charset;
+	ds2do(&ses->ds, &o);
 	if ((o.plain = l ? l->plain : 1) == -1) o.plain = 0;
 	if (l) l->plain = o.plain;
 	memcpy(&o.default_fg, &default_fg, sizeof(struct rgb));
@@ -966,7 +1220,30 @@ void html_interpret(struct session *ses)
 	memcpy(&o.default_link, &default_link, sizeof(struct rgb));
 	memcpy(&o.default_vlink, &default_vlink, sizeof(struct rgb));
 	o.framename = "";
+	foreach(fd, ses->scrn_frames) fd->used = 0;
 	cached_format_html(l, ses->screen, &o);
+	if (ses->screen->f_data && ses->screen->f_data->frame) {
+		cf = current_frame(ses);
+		format_frames(ses, ses->screen->f_data->frame_desc, &o, 0);
+	}
+	foreach(fd, ses->scrn_frames) if (!fd->used) {
+		struct f_data_c *fdp = fd->prev;
+		detach_formatted(fd);
+		del_from_list(fd);
+		mem_free(fd);
+		fd = fdp;
+	}
+	if (cf) {
+		int n = 0;
+		foreach(fd, ses->scrn_frames) {
+			if (fd->f_data && fd->f_data->frame) continue;
+			if (fd == cf) {
+				cur_loc(ses)->vs.current_link = n;
+				break;
+			}
+			n++;
+		}
+	}
 }
 
 #define SRCH_ALLOC_GR	0x10000
@@ -975,11 +1252,6 @@ void add_srch_chr(struct f_data *f, unsigned char c, int x, int y, int nn)
 {
 	int n = f->nsearch;
 	if (c == ' ' && (!n || f->search[n - 1].c == ' ')) return;
-	if (!(f->nsearch & (SRCH_ALLOC_GR - 1))) {
-		struct search *s;
-		if (!(s = mem_realloc(f->search, (f->nsearch + SRCH_ALLOC_GR) * sizeof(struct search)))) return;
-		f->search = s;
-	}
 	f->search[n].c = c;
 	f->search[n].x = x;
 	f->search[n].y = y;
@@ -1022,7 +1294,6 @@ void sort_srch(struct f_data *f)
 	memset(f->slines2, 0, f->y * sizeof(struct search *));
 	for (i = 0; i < f->y; i++) min[i] = MAXINT, max[i] = 0;
 	for (i = 0; i < f->nsearch; i++) {
-		int p, q, j;
 		struct search *s = &f->search[i];
 		if (s->x < min[s->y]) min[s->y] = s->x, f->slines1[s->y] = s;
 		if (s->x + s->n > max[s->y]) max[s->y] = s->x + s->n, f->slines2[s->y] = s;
@@ -1031,11 +1302,11 @@ void sort_srch(struct f_data *f)
 	mem_free(max);
 }
 
-void get_search_data(struct f_data *f)
+int get_srch(struct f_data *f)
 {
 	struct node *n;
-	if (f->search) return;
-	f->search = DUMMY; f->nsearch = 0;
+	int cnt = 0;
+	int cc = !f->search;
 	foreachback(n, f->nodes) {
 		int x, y;
 		int xm = n->x + n->xw, ym = n->y + n->yw;
@@ -1048,24 +1319,40 @@ void get_search_data(struct f_data *f)
 				if (c < ' ') c = ' ' ;
 				if (c == ' ' && ns) continue;
 				if (ns) {
-					add_srch_chr(f, c, x, y, 1);
+					if (!cc) add_srch_chr(f, c, x, y, 1);
+					else cnt++;
 					ns = 0;
 					continue;
 				}
-				if (c != ' ') add_srch_chr(f, c, x, y, 1);
+				if (c != ' ')  if (!cc) add_srch_chr(f, c, x, y, 1);
+					       else cnt++;
 				else {
 					int xx;
 					for (xx = x + 1; xx < xm && xx < f->data[y].l; xx++) if ((unsigned char) f->data[y].d[xx] >= ' ') goto ja_uz_z_toho_programovani_asi_zcvoknu;
 					xx = x;
 					ja_uz_z_toho_programovani_asi_zcvoknu:
-					add_srch_chr(f, ' ', x, y, xx - x);
+					if (!cc) add_srch_chr(f, ' ', x, y, xx - x);
+					else cnt++;
 					if (xx == x) break;
 					x = xx - 1;
 				}
 			}
-			add_srch_chr(f, ' ', x, y, 0);
+			if (!cc) add_srch_chr(f, ' ', x, y, 0);
+			else cnt++;
 		}
 	}
+	return cnt;
+	
+}
+
+void get_search_data(struct f_data *f)
+{
+	int n;
+	if (f->search) return;
+	n = get_srch(f);
+	f->nsearch = 0;
+	if (!(f->search = mem_alloc(n * sizeof(struct search)))) return;
+	get_srch(f);
 	while (f->nsearch && f->search[f->nsearch - 1].c == ' ') f->nsearch--;
 	/*debug("!");*/
 	sort_srch(f);
