@@ -89,6 +89,7 @@ struct table_cell {
 	int x_width;
 	int height;
 	int xpos, ypos, xw, yw;
+	int link_num;
 };
 
 struct table_column {
@@ -115,6 +116,7 @@ struct table {
 	int *xcols;
 	int *r_heights;
 	int rh;
+	int link_num;
 };
 
 #define CELL(t, x, y) (&(t)->cells[(y) * (t)->rx + (x)])
@@ -308,7 +310,7 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	struct table_cell *cell;
 	unsigned char *t_name, *t_attr, *en;
 	int t_namelen;
-	int x, y = -1;
+	int x = 0, y = -1;
 	int p = 0;
 	unsigned char *lbhp = NULL;
 	int l_al = AL_LEFT;
@@ -318,7 +320,6 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	int i, j, k;
 	struct rgb l_col;
 	int c_al = AL_TR, c_val = VAL_TR, c_width = W_AUTO, c_span = 0;
-	struct table_column *col;
 	memcpy(&l_col, bgcolor, sizeof(struct rgb));
 	*end = html;
 	if (bad_html) {
@@ -404,7 +405,6 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
 	}
 	if (t_namelen == 2 && !casecmp(t_name, "TR", 2)) {
-		char *al;
 		if (c_span) new_columns(t, c_span, c_width, c_al, c_val, 1);
 		if (p) CELL(t, x, y)->end = html, p = 0;
 		if (lbhp) (*bad_html)[*bhp-1].e = html, lbhp = NULL;
@@ -519,14 +519,16 @@ struct table *parse_table(unsigned char *html, unsigned char *eof, unsigned char
 	return t;
 }
 
-void get_cell_width(char *start, char *end, int cellpd, int w, int a, int *min, int *max)
+void get_cell_width(char *start, char *end, int cellpd, int w, int a, int *min, int *max, int n_link, int *n_links)
 {
 	struct part *p;
 	if (min) *min = -1;
 	if (max) *max = -1;
-	if (!(p = format_html_part(start, end, AL_LEFT, cellpd, w, NULL, !!a, !!a, NULL))) return;
+	if (n_links) *n_links = n_link;
+	if (!(p = format_html_part(start, end, AL_LEFT, cellpd, w, NULL, !!a, !!a, NULL, n_link))) return;
 	if (min) *min = p->x;
 	if (max) *max = p->xmax;
+	if (n_links) *n_links = p->link_num;
 	if (min && max && *min > *max) internal("get_cell_width: %d > %d", *min, *max);
 	mem_free(p);
 }
@@ -535,25 +537,31 @@ static inline void check_cell_widths(struct table *t)
 {
 	int i, j;
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
-		int min, max;
+		int min, max, nl;
 		struct table_cell *c = CELL(t, i, j);
 		if (!c->start) continue;
-		get_cell_width(c->start, c->end, t->cellpd, 0, 0, &min, &max);
+		get_cell_width(c->start, c->end, t->cellpd, 0, 0, &min, &max, c->link_num, NULL);
 		if (min != c->min_width || max < c->max_width) internal("check_cell_widths failed");
 	}
 }
 
+#define g_c_w(cc)							\
+do {									\
+		struct table_cell *c = cc;				\
+		if (!c->start) continue;				\
+		c->link_num = nl;					\
+		get_cell_width(c->start, c->end, t->cellpd, 0, 0, &c->min_width, &c->max_width, nl, &nl);\
+} while (0)
+
 void get_cell_widths(struct table *t)
 {
+	int nl = t->p->link_num;
 	int i, j;
-	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
-		struct table_cell *c = CELL(t, i, j);
-		if (!c->start) continue;
-		get_cell_width(c->start, c->end, t->cellpd, 0, 0, &c->min_width, &c->max_width);
-		/*if (i < t->xc && t->xcols[i] >= 0) {
-			if (t->xcols[i] > c->max_width) c->max_width = t->xcols[i];
-		}*/
-	}
+	if (!d_opt->table_order)
+		for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) g_c_w(CELL(t, i, j));
+	else
+		for (i = 0; i < t->x; i++) for (j = 0; j < t->y; j++) g_c_w(CELL(t, i, j));
+	t->link_num = nl;
 }
 
 void dst_width(int *p, int n, int w, int *lim)
@@ -786,7 +794,7 @@ void check_table_widths(struct table *t)
 	int *w;
 	int i, j;
 	int s, ns;
-	int m, mi;
+	int m, mi = 0; /* go away, warning! */
 	if (!(w = mem_alloc(t->x * sizeof(int)))) return;
 	memset(w, 0, t->x * sizeof(int));
 	for (j = 0; j < t->y; j++) for (i = 0; i < t->x; i++) {
@@ -795,10 +803,10 @@ void check_table_widths(struct table *t)
 		if (!c->start) continue;
 		for (k = 1; k < c->colspan; k++) p += get_vline_width(t, i + k) >= 0;
 		for (k = 0; k < c->colspan; k++) p += t->w_c[i + k];
-		get_cell_width(c->start, c->end, t->cellpd, p, 1, &c->x_width, NULL);
+		get_cell_width(c->start, c->end, t->cellpd, p, 1, &c->x_width, NULL, c->link_num, NULL);
 		if (c->x_width > p) {
 			int min, max;
-			get_cell_width(c->start, c->end, t->cellpd, 0, 0, &min, &max);
+			get_cell_width(c->start, c->end, t->cellpd, 0, 0, &min, &max, c->link_num, NULL);
 			internal("cell is now wider (%d > %d) min = %d, max = %d, now_min = %d, now_max = %d", c->x_width, p, t->min_c[i], t->max_c[i], min, max);
 		}
 	}
@@ -878,7 +886,7 @@ void get_table_heights(struct table *t)
 				xw += t->w_c[i + sp];
 				if (sp < cell->colspan - 1) xw += get_vline_width(t, i + sp + 1) >= 0;
 			}
-			if (!(p = format_html_part(cell->start, cell->end, cell->align, t->cellpd, xw, NULL, 2, 2, NULL))) return;
+			if (!(p = format_html_part(cell->start, cell->end, cell->align, t->cellpd, xw, NULL, 2, 2, NULL, cell->link_num))) return;
 			cell->height = p->y;
 				/*debug("%d, %d.",xw, cell->height);*/
 			mem_free(p);
@@ -932,7 +940,7 @@ void display_complicated_table(struct table *t, int x, int y, int *yy)
 				if (cell->b) format.attr |= AT_BOLD;
 				memcpy(&format.bg, &cell->bgcolor, sizeof(struct rgb));
 				memcpy(&par_format.bgcolor, &cell->bgcolor, sizeof(struct rgb));
-				p = format_html_part(cell->start, cell->end, cell->align, t->cellpd, xw, f, t->p->xp + xp, t->p->yp + yp + (cell->valign != VAL_MIDDLE && cell->valign != VAL_BOTTOM ? 0 : (yw - cell->height) / (cell->valign == VAL_MIDDLE ? 2 : 1)), NULL);
+				p = format_html_part(cell->start, cell->end, cell->align, t->cellpd, xw, f, t->p->xp + xp, t->p->yp + yp + (cell->valign != VAL_MIDDLE && cell->valign != VAL_BOTTOM ? 0 : (yw - cell->height) / (cell->valign == VAL_MIDDLE ? 2 : 1)), NULL, cell->link_num);
 				cell->xpos = xp;
 				cell->ypos = yp;
 				cell->xw = xw;
@@ -1022,8 +1030,6 @@ void display_table_frames(struct table *t, int x, int y)
 
 	cy = y;
 	for (j = 0; j <= t->y; j++) {
-		int cy1 = cy;
-		int cyy;
 		cx = x;
 		if ((j > 0 && j < t->y && get_hline_width(t, j) >= 0) || (j == 0 && t->border && (t->frame & F_ABOVE)) || (j == t->y && t->border && (t->frame & F_BELOW))) {
 			for (i = 0; i < t->x; i++) {
@@ -1066,14 +1072,13 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	struct part *p = f;
 	int border, cellsp, vcellpd, cellpd, align;
 	int frame, rules, width, wf;
-	int *min, *max;
 	struct rgb bgcolor;
 	struct table *t;
 	char *al;
-	int cye, y;
+	int cye;
 	int x;
 	int i;
-	int llm = last_link_to_move;
+	/*int llm = last_link_to_move;*/
 	struct s_e *bad_html;
 	int bad_html_n;
 	struct node *n, *nn;
@@ -1218,11 +1223,12 @@ void format_table(unsigned char *attr, unsigned char *html, unsigned char *eof,
 	p->cy = cye;
 
 	ret2:
+	p->link_num = t->link_num;
 	if (p->cy > p->y) p->y = p->cy;
-	ret1:
+	/*ret1:*/
 	free_table(t);
 	ret0:
 	kill_html_stack_item(&html_top);
-	ret:
+	/*ret:*/
 	table_level--;
 }
