@@ -1,6 +1,7 @@
 #include "links.h"
 
 struct http_connection_info {
+	int bl_flags;
 	int http10;
 	int close;
 	int length;
@@ -8,7 +9,7 @@ struct http_connection_info {
 	int chunk_remaining;
 };
 
-unsigned char *parse_http_header(unsigned char *head, unsigned char *item)
+unsigned char *parse_http_header(unsigned char *head, unsigned char *item, unsigned char **ptr)
 {
 	unsigned char *i, *f, *g, *h;
 	if (!head) return NULL;
@@ -20,12 +21,17 @@ unsigned char *parse_http_header(unsigned char *head, unsigned char *item)
 			if (upcase(*i) != upcase(*f)) goto cont;
 		if (!*f) break;
 		if (f[0] == ':') {
-			if (f[1] == ' ') f++;
+			while (f[1] == ' ') f++;
 			for (g = ++f; *g >= ' '; g++) ;
+			while (g > f && g[-1] == ' ') g--;
 			if (h) mem_free(h);
 			if ((h = mem_alloc(g - f + 1))) {
 				memcpy(h, f, g - f);
 				h[g - f] = 0;
+				if (ptr) {
+					*ptr = f;
+					break;
+				}
 			}
 		}
 		cont:;
@@ -48,12 +54,13 @@ unsigned char *parse_header_param(unsigned char *x, unsigned char *e)
 	while (*y && (*y <= ' ' || *y == '=')) y++;
 	if (!*y) return stracpy("");
 	lp = 0;
-	while (y[lp] > ' ') lp++;
+	while (y[lp] >= ' ' && y[lp] != ';') lp++;
 	return memacpy(y, lp);
 }
 
 static int get_http_code(unsigned char *head, int *code, int *version)
 {
+	while (head[0] == ' ') head++;
 	if (upcase(head[0]) != 'H' || upcase(head[1]) != 'T' || upcase(head[2]) != 'T' ||
 	    upcase(head[3]) != 'P') return -1;
 	if (head[4] == '/' && head[5] >= '0' && head[5] <= '9'
@@ -68,13 +75,13 @@ static int get_http_code(unsigned char *head, int *code, int *version)
 	return 0;
 }
 
-unsigned char *buggy_servers[] = { "mod_czech/3.1.0", "Purveyor", NULL };
+unsigned char *buggy_servers[] = { "mod_czech/3.1.0", "Purveyor", "Netscape-Enterprise", NULL };
 
 int check_http_server_bugs(unsigned char *url, struct http_connection_info *info, unsigned char *head)
 {
 	unsigned char *server, **s;
-	if (info->http10) return 0;
-	if (!(server = parse_http_header(head, "Server"))) return 0;
+	if (!http_bugs.allow_blacklist || info->http10) return 0;
+	if (!(server = parse_http_header(head, "Server", NULL))) return 0;
 	for (s = buggy_servers; *s; s++) if (strstr(server, *s)) goto bug;
 	mem_free(server);
 	return 0;
@@ -92,11 +99,11 @@ void http_end_request(struct connection *c)
 {
 	if (c->state == S_OK) {
 		if (c->cache) {
-			truncate_entry(c->cache, c->from);
+			truncate_entry(c->cache, c->from, 1);
 			c->cache->incomplete = 0;
 		}
 	}
-	if (c->info && !((struct http_connection_info *)c->info)->close && (!bug_post_no_keepalive || !strchr(c->url, POST_CHAR))) {
+	if (c->info && !((struct http_connection_info *)c->info)->close && (!http_bugs.bug_post_no_keepalive || !strchr(c->url, POST_CHAR))) {
 		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT);
 	} else abort_connection(c);
 }
@@ -127,13 +134,15 @@ void http_get_header(struct connection *);
 
 void http_send_header(struct connection *c)
 {
+	static unsigned char *accept_charset = NULL;
 	struct http_connection_info *info;
-	int http10 = 0;
+	int http10 = http_bugs.http10;
 	struct cache_entry *e = NULL;
 	unsigned char *hdr;
 	unsigned char *h, *u;
 	int l = 0;
 	unsigned char *post;
+	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
 	set_timeout(c);
 	if (!(info = mem_alloc(sizeof(struct http_connection_info)))) {
 		setcstate(c, S_OUT_OF_MEM);
@@ -142,10 +151,11 @@ void http_send_header(struct connection *c)
 	}
 	memset(info, 0, sizeof(struct http_connection_info));
 	c->info = info;
-	if ((h = get_host_name(c->url))) {
-		if (get_blacklist_flags(h) & BL_HTTP10) http10 = 1;
+	if ((h = get_host_name(host))) {
+		info->bl_flags = get_blacklist_flags(h);
 		mem_free(h);
 	}
+	if (info->bl_flags & BL_HTTP10) http10 = 1;
 	info->http10 = http10;
 	post = strchr(c->url, POST_CHAR);
 	if (post) post++;
@@ -154,7 +164,6 @@ void http_send_header(struct connection *c)
 		http_end_request(c);
 		return;
 	}
-	setcstate(c, S_SENT);
 	if (!post) add_to_str(&hdr, &l, "GET ");
 	else {
 		add_to_str(&hdr, &l, "POST ");
@@ -170,7 +179,7 @@ void http_send_header(struct connection *c)
 	else add_bytes_to_str(&hdr, &l, u, post - u - 1);
 	if (!http10) add_to_str(&hdr, &l, " HTTP/1.1\r\n");
 	else add_to_str(&hdr, &l, " HTTP/1.0\r\n");
-	if ((h = get_host_name(upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url)))) {
+	if ((h = get_host_name(host))) {
 		add_to_str(&hdr, &l, "Host: ");
 		add_to_str(&hdr, &l, h);
 		mem_free(h);
@@ -179,10 +188,27 @@ void http_send_header(struct connection *c)
 	add_to_str(&hdr, &l, "User-Agent: Links (" VERSION_STRING "; ");
 	add_to_str(&hdr, &l, system_name);
 	add_to_str(&hdr, &l, ")\r\n");
+	add_to_str(&hdr, &l, "Accept: */*\r\n");
+	if (!(accept_charset)) {
+		int i;
+		unsigned char *cs, *ac;
+		int aclen = 0;
+		ac = init_str();
+		for (i = 0; (cs = get_cp_mime_name(i)); i++) {
+			if (aclen) add_to_str(&ac, &aclen, ", ");
+			else add_to_str(&ac, &aclen, "Accept-Charset: ");
+			add_to_str(&ac, &aclen, cs);
+		}
+		if (aclen) add_to_str(&ac, &aclen, "\r\n");
+		if ((accept_charset = malloc(strlen(ac) + 1))) strcpy(accept_charset, ac);
+		else accept_charset = "";
+		mem_free(ac);
+	}
+	if (!(info->bl_flags & BL_NO_CHARSET)) add_to_str(&hdr, &l, accept_charset);
 	if (!http10) {
 		if (upcase(c->url[0]) != 'P') add_to_str(&hdr, &l, "Connection: ");
 		else add_to_str(&hdr, &l, "Proxy-Connection: ");
-		if (!post || !bug_post_no_keepalive) add_to_str(&hdr, &l, "Keep-Alive\r\n");
+		if (!post || !http_bugs.bug_post_no_keepalive) add_to_str(&hdr, &l, "Keep-Alive\r\n");
 		else add_to_str(&hdr, &l, "close\r\n");
 	}
 	if ((e = c->cache)) {
@@ -211,6 +237,7 @@ void http_send_header(struct connection *c)
 		add_num_to_str(&hdr, &l, strlen(post) / 2);
 		add_to_str(&hdr, &l, "\r\n");
 	}
+	send_cookies(&hdr, &l, host);
 	add_to_str(&hdr, &l, "\r\n");
 	if (post) {
 		while (post[0] && post[1]) {
@@ -225,6 +252,7 @@ void http_send_header(struct connection *c)
 	}
 	write_to_socket(c, c->sock1, hdr, strlen(hdr), http_get_header);
 	mem_free(hdr);
+	setcstate(c, S_SENT);
 }
 
 int is_line_in_buffer(struct read_buffer *rb)
@@ -248,7 +276,6 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 		http_end_request(c);
 		return;
 	}
-	setcstate(c, S_TRANS);
 	if (info->length != -2) {
 		int l = rb->len;
 		if (info->length >= 0 && info->length < l) l = info->length;
@@ -322,6 +349,7 @@ void read_http_data(struct connection *c, struct read_buffer *rb)
 	}
 	read_more:
 	read_from_socket(c, c->sock1, rb, read_http_data);
+	setcstate(c, S_TRANS);
 }
 
 int get_header(struct read_buffer *rb)
@@ -329,7 +357,7 @@ int get_header(struct read_buffer *rb)
 	int i;
 	for (i = 0; i < rb->len; i++) {
 		unsigned char a = rb->data[i];
-		if (a < ' ' && a != 10 && a != 13) return -1;
+		if (/*a < ' ' && a != 10 && a != 13*/!a) return -1;
 		if (i < rb->len - 1 && a == 10 && rb->data[i + 1] == 10) return i + 2;
 		if (i < rb->len - 3 && a == 13) {
 			if (rb->data[i + 1] != 10) return -1;
@@ -344,13 +372,32 @@ int get_header(struct read_buffer *rb)
 
 void http_got_header(struct connection *c, struct read_buffer *rb)
 {
+	int state = c->state != S_PROC ? S_GETH : S_PROC;
 	unsigned char *head;
+	unsigned char *cookie, *ch;
 	int a, h, version;
 	unsigned char *d;
 	struct cache_entry *e;
 	struct http_connection_info *info;
+	unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
 	set_timeout(c);
-	setcstate(c, S_GETH);
+	info = c->info;
+	if (rb->close == 2) {
+		unsigned char *h;
+		if (!c->tries && (h = get_host_name(host))) {
+			if (info->bl_flags & BL_NO_CHARSET) {
+				del_blacklist_entry(h, BL_NO_CHARSET);
+			} else {
+				add_blacklist_entry(h, BL_NO_CHARSET);
+				c->tries = -1;
+			}
+			mem_free(h);
+		}
+		setcstate(c, S_CANT_READ);
+		retry_connection(c);
+		return;
+	}
+	rb->close = 0;
 	again:
 	if ((a = get_header(rb)) == -1) {
 		setcstate(c, S_HTTP_ERROR);
@@ -359,6 +406,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	if (!a) {
 		read_from_socket(c, c->sock1, rb, http_got_header);
+		setcstate(c, state);
 		return;
 	}
 	if (get_http_code(rb->data, &h, &version) || h == 101) {
@@ -372,15 +420,21 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	memcpy(head, rb->data, a); head[a] = 0;
-	if (check_http_server_bugs(c->url, c->info, head)) {
+	if (check_http_server_bugs(host, c->info, head)) {
 		mem_free(head);
 		setcstate(c, S_RESTART);
 		retry_connection(c);
 		return;
 	}
+	ch = head;
+	while ((cookie = parse_http_header(ch, "Set-Cookie", &ch))) {
+		unsigned char *host = upcase(c->url[0]) != 'P' ? c->url : get_url_data(c->url);
+		set_cookie(NULL, host, cookie);
+		mem_free(cookie);
+	}
 	if (h == 100) {
 		mem_free(head);
-		setcstate(c, S_PROC);
+		state = S_PROC;
 		kill_buffer_data(rb, a);
 		goto again;
 	}
@@ -410,7 +464,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	if (h == 301 || h == 302 || h == 303) {
-		if ((d = parse_http_header(e->head, "Location"))) {
+		if ((d = parse_http_header(e->head, "Location", NULL))) {
 			if (e->redirect) mem_free(e->redirect);
 			e->redirect = d;
 			e->redirect_get = h == 303;
@@ -418,16 +472,15 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 	}
 	kill_buffer_data(rb, a);
 	c->cache = e;
-	info = c->info;
 	info->close = 0;
 	info->length = -1;
 	info->version = version;
-	if ((d = parse_http_header(e->head, "Connection")) || (d = parse_http_header(e->head, "Proxy-Connection"))) {
+	if ((d = parse_http_header(e->head, "Connection", NULL)) || (d = parse_http_header(e->head, "Proxy-Connection", NULL))) {
 		if (!strcasecmp(d, "close")) info->close = 1;
 		mem_free(d);
 	} else if (version < 11) info->close = 1;
 	c->from = 0;
-	if ((d = parse_http_header(e->head, "Content-Range"))) {
+	if ((d = parse_http_header(e->head, "Content-Range", NULL))) {
 		if (strlen(d) > 6) {
 			d[5] = 0;
 			if (!(strcasecmp(d, "bytes")) && d[6] >= '0' && d[6] <= '9') {
@@ -437,7 +490,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		}
 		mem_free(d);
 	}
-	if ((d = parse_http_header(e->head, "Content-Length"))) {
+	if ((d = parse_http_header(e->head, "Content-Length", NULL))) {
 		unsigned char *ep;
 		int l = strtol(d, (char **)&ep, 10);
 		if (!*ep && l >= 0) {
@@ -446,16 +499,16 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		}
 		mem_free(d);
 	}
-	if ((d = parse_http_header(e->head, "Accept-Ranges"))) mem_free(d);
+	if ((d = parse_http_header(e->head, "Accept-Ranges", NULL))) mem_free(d);
 	else if (!c->unrestartable) c->unrestartable = 1;
-	if ((d = parse_http_header(e->head, "Transfer-Encoding"))) {
+	if ((d = parse_http_header(e->head, "Transfer-Encoding", NULL))) {
 		if (!strcasecmp(d, "chunked")) {
 			info->length = -2;
 			info->chunk_remaining = -1;
 		}
 		mem_free(d);
 	}
-	if ((d = parse_http_header(e->head, "Last-Modified"))) {
+	if ((d = parse_http_header(e->head, "Last-Modified", NULL))) {
 		if (e->last_modified && strcasecmp(e->last_modified, d)) {
 			delete_entry_content(e);
 			if (c->from) {
@@ -469,7 +522,7 @@ void http_got_header(struct connection *c, struct read_buffer *rb)
 		if (!e->last_modified) e->last_modified = d;
 		else mem_free(d);
 	}
-	if (!e->last_modified && (d = parse_http_header(e->head, "Date")))
+	if (!e->last_modified && (d = parse_http_header(e->head, "Date", NULL)))
 		e->last_modified = d;
 	if (info->length == -1 || (version < 11 && info->close)) rb->close = 1;
 	read_http_data(c, rb);
@@ -480,5 +533,6 @@ void http_get_header(struct connection *c)
 	struct read_buffer *rb;
 	set_timeout(c);
 	if (!(rb = alloc_read_buffer(c))) return;
+	rb->close = 1;
 	read_from_socket(c, c->sock1, rb, http_got_header);
 }
