@@ -1,5 +1,7 @@
 #include "links.h"
 
+tcount connection_count = 0;
+
 int active_connections = 0;
 
 struct list_head queue = {&queue, &queue};
@@ -28,7 +30,7 @@ struct list_head keepalive_connections = {&keepalive_connections, &keepalive_con
 
 long connect_info(int type)
 {
-	int i = 0, j;
+	int i = 0;
 	struct connection *ce;
 	struct k_conn *cee;
 	switch (type) {
@@ -51,6 +53,14 @@ long connect_info(int type)
 	}
 	return 0;
 }
+
+int connection_disappeared(struct connection *c, tcount count)
+{
+	struct connection *d;
+	foreach(d, queue) if (c == d && count == d->count) return 0;
+	return 1;
+}
+
 struct h_conn *is_host_on_list(struct connection *c)
 {
 	unsigned char *ho;
@@ -96,6 +106,7 @@ void setcstate(struct connection *c, int state)
 	if ((c->state = state) == S_TRANS) {
 		struct remaining_info *r = &c->prg;
 		if (r->timer == -1) {
+			tcount count = c->count;
 			if (!r->valid) {
 				memset(r, 0, sizeof(struct remaining_info));
 				r->valid = 1;
@@ -105,6 +116,7 @@ void setcstate(struct connection *c, int state)
 			st_r = 1;
 			stat_timer(c);
 			st_r = 0;
+			if (connection_disappeared(c, count)) return;
 		}
 	} else {
 		struct remaining_info *r = &c->prg;
@@ -188,15 +200,18 @@ void free_connection_data(struct connection *c)
 			mem_free(h->host);
 			mem_free(h);
 		}
-	} else internal("suspending connection that is not on the list");
+	} else internal("suspending connection that is not on the list (state %d)", c->state);
 }
 
 void send_connection_info(struct connection *c)
 {
+	tcount count = c->count;
 	struct status *stat = c->statuss.next;
 	while ((void *)stat != &c->statuss) {
+		stat->ce = c->cache;
 		stat = stat->next;
 		if (stat->prev->end) stat->prev->end(stat->prev, stat->prev->data);
+		if (connection_disappeared(c, count)) return;
 	}
 }
 
@@ -208,7 +223,6 @@ void del_connection(struct connection *c)
 	mem_free(c);
 }
 
-void check_queue();
 #ifdef DEBUG
 void check_queue_bugs();
 #endif
@@ -393,7 +407,7 @@ void retry_connection(struct connection *c)
 
 void abort_connection(struct connection *c)
 {
-	interrupt_connection(c);
+	if (c->running) interrupt_connection(c);
 	/*send_connection_info(c);*/
 	del_connection(c);
 #ifdef DEBUG
@@ -402,18 +416,20 @@ void abort_connection(struct connection *c)
 	register_bottom_half((void (*)(void *))check_queue, NULL);
 }
 
-void try_connection(struct connection *c)
+int try_connection(struct connection *c)
 {
 	struct h_conn *hc = NULL;
 	if ((hc = is_host_on_list(c))) {
 		if (hc->conn >= max_connections_to_host)
-			if (try_to_suspend_connection(c, hc->host)) return;
-			else goto connect_ok;
+			if (try_to_suspend_connection(c, hc->host)) return 0;
+			else return -1;
 	}
 	if (active_connections >= max_connections)
-		if (try_to_suspend_connection(c, NULL)) return;
+		if (try_to_suspend_connection(c, NULL)) return 0;
+		else return -1;
 	connect_ok:
 	run_connection(c);
+	return 1;
 }
 
 #ifdef DEBUG
@@ -451,7 +467,9 @@ void check_queue_bugs()
 
 void check_queue()
 {
-	struct connection *c = queue.next;
+	struct connection *c;
+	again:
+	c = queue.next;
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
@@ -461,21 +479,25 @@ void check_queue()
 		int cp = getpri(c);
 		for (d = c; d != (struct connection *)&queue && getpri(d) == cp;) {
 			struct connection *dd = d; d = d->next;
-			if (!dd->state) if (is_host_on_keepalive_list(dd))
-				try_connection(dd);
+			if (!dd->state) if (is_host_on_keepalive_list(dd)) {
+				if (try_connection(dd)) goto again;
+			}
 		}
 		for (d = c; d != (struct connection *)&queue && getpri(d) == cp;) {
 			struct connection *dd = d; d = d->next;
-			if (!dd->state) try_connection(dd);
+			if (!dd->state) {
+				if (try_connection(dd)) goto again;
+			}
 		}
 		c = d;
 	}
+	again2:
 	foreachback(c, queue) {
 		if (getpri(c) < PRI_CANCEL) break;
 		if (c->state == S_WAIT) {
-			c = c->prev;
-			setcstate(c->next, S_INTERRUPTED);
-			del_connection(c->next);
+			setcstate(c, S_INTERRUPTED);
+			del_connection(c);
+			goto again2;
 		}
 	}
 #ifdef DEBUG
@@ -502,6 +524,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	struct cache_entry *e = NULL;
 	struct connection *c;
 	unsigned char *u;
+	if (stat) stat->c = NULL, stat->ce = NULL, stat->pri = pri;
 #ifdef DEBUG
 	foreach(c, queue) {
 		struct status *st;
@@ -518,6 +541,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	if (stat) stat->state = S_OUT_OF_MEM, stat->prev_error = 0;
 	if (no_cache <= NC_CACHE && !find_in_cache(url, &e) && !e->incomplete) {
 		if (stat) {
+			stat->ce = e;
 			stat->state = S_OK;
 			if (stat->end) stat->end(stat, stat->data);
 		}
@@ -528,7 +552,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 		mem_free(c);
 		return -1;
 	}
-	foreach(c, queue) if (!strcmp(c->url, u)) {
+	foreach(c, queue) if (!c->detached && !strcmp(c->url, u)) {
 		mem_free(u);
 		if (getpri(c) > pri) {
 			del_from_list(c);
@@ -538,6 +562,8 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 		} else c->pri[pri]++;
 		if (stat) {
 			stat->prg = &c->prg;
+			stat->c = c;
+			stat->ce = c->cache;
 			add_to_list(c->statuss, stat);
 			setcstate(c, c->state);
 		}
@@ -552,6 +578,7 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 		return -1;
 	}
 	memset(c, 0, sizeof(struct connection));
+	c->count = connection_count++;
 	c->url = u;
 	c->running = 0;
 	c->prev_error = 0;
@@ -572,6 +599,8 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	c->timer = -1;
 	if (stat) {
 		stat->prg = &c->prg;
+		stat->c = c;
+		stat->ce = NULL;
 		add_to_list(c->statuss, stat);
 	}
 	add_to_queue(c);
@@ -583,64 +612,95 @@ int load_url(unsigned char *url, struct status *stat, int pri, int no_cache)
 	return 0;
 }
 
-void change_connection(unsigned char *url, struct status *oldstat, struct status *newstat, int oldpri, int newpri)
+void change_connection(struct status *oldstat, struct status *newstat, int newpri)
 {		/* !!! FIXME: one object in more connections */
 	struct connection *c;
-	unsigned char *u;
-	int a = 0;
-	if (oldstat && oldstat->state < 0) {
+	int oldpri;
+	if (!oldstat) {
+		internal("change_connection: oldstat == NULL");
+		return;
+	}
+	oldpri = oldstat->pri;
+	if (oldstat->state < 0) {
 		if (newstat) {
+			newstat->ce = oldstat->ce;
 			newstat->state = oldstat->state;
 			newstat->prev_error = oldstat->prev_error;
 			if (newstat->end) newstat->end(newstat, newstat->data);
 		}
 		return;
 	}
-	if (!(u = get_proxy(url))) return;
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
-	foreach(c, queue) if (!strcmp(c->url, u)) {
-		if (--c->pri[oldpri] < 0) {
-			internal("priority counter underflow");
-			c->pri[oldpri] = 0;
-		}
-		c->pri[newpri]++;
-		if (oldstat) {
-			del_from_list(oldstat);
-			oldstat->state = S_INTERRUPTED;
-		}
-		if (newstat) {
-			newstat->prg = &c->prg;
-			add_to_list(c->statuss, newstat);
-			newstat->state = c->state;
-			newstat->prev_error = c->prev_error;
-		}
-		a = 1;
+	c = oldstat->c;
+	if (--c->pri[oldpri] < 0) {
+		internal("priority counter underflow");
+		c->pri[oldpri] = 0;
+	}
+	c->pri[newpri]++;
+	del_from_list(oldstat);
+	oldstat->state = S_INTERRUPTED;
+	if (newstat) {
+		newstat->prg = &c->prg;
+		add_to_list(c->statuss, newstat);
+		newstat->state = c->state;
+		newstat->prev_error = c->prev_error;
+		newstat->pri = newpri;
+		newstat->c = c;
+		newstat->ce = c->cache;
+	}
+	if (c->detached && !newstat) {
+		setcstate(c, S_INTERRUPTED);
+		abort_connection(c);
 	}
 	sort_queue();
-	if (!a && newstat) {
-		newstat->state = S_OK;
-		newstat->prev_error = 0;
-	}
-	mem_free(u);
 #ifdef DEBUG
 	check_queue_bugs();
 #endif
 	register_bottom_half((void (*)(void *))check_queue, NULL);
 }
 
+void detach_connection(struct status *stat, int pos)
+{
+	struct connection *c;
+	int i, l;
+	if (stat->state < 0) return;
+	c = stat->c;
+	if (c->detached) goto detach_done;
+	if (!c->cache) return;
+	if (c->est_length == -1) l = c->from;
+	else l = c->est_length;
+	if (l < memory_cache_size * MAX_CACHED_OBJECT) return;
+	l = 0;
+	for (i = 0; i < PRI_CANCEL; i++) l += c->pri[i];
+	if (!l) internal("detaching free connection");
+	delete_unused_format_cache_entries();
+	if (l != 1 || c->cache->refcount) return;
+	/*debug("detached");*/
+	c->cache->url[0] = 0;
+	c->detached = 1;
+	detach_done:
+	free_entry_to(c->cache, pos);
+}
+
 void connection_timeout(struct connection *c)
 {
 	c->timer = -1;
 	setcstate(c, S_TIMEOUT);
-	retry_connection(c);
+	if (c->dnsquery) abort_connection(c);
+	else retry_connection(c);
+}
+
+void connection_timeout_1(struct connection *c)
+{
+	c->timer = install_timer((c->unrestartable ? unrestartable_receive_timeout : receive_timeout) * 500, (void (*)(void *))connection_timeout, c);
 }
 
 void set_timeout(struct connection *c)
 {
 	if (c->timer != -1) kill_timer(c->timer);
-	c->timer = install_timer((c->unrestartable ? unrestartable_receive_timeout : receive_timeout) * 1000, (void (*)(void *))connection_timeout, c);
+	c->timer = install_timer((c->unrestartable ? unrestartable_receive_timeout : receive_timeout) * 500, (void (*)(void *))connection_timeout_1, c);
 }
 
 void reset_timeout(struct connection *c)
@@ -700,6 +760,19 @@ void add_blacklist_entry(unsigned char *host, int flags)
 	add_to_list(blacklist, b);
 }
 
+void del_blacklist_entry(unsigned char *host, int flags)
+{
+	struct blacklist_entry *b;
+	foreach(b, blacklist) if (!strcasecmp(host, b->host)) {
+		b->flags &= ~flags;
+		if (!b->flags) {
+			del_from_list(b);
+			mem_free(b);
+		}
+		return;
+	}
+}
+
 int get_blacklist_flags(unsigned char *host)
 {
 	struct blacklist_entry *b;
@@ -713,41 +786,42 @@ void free_blacklist()
 }
 
 struct s_msg_dsc msg_dsc[] = {
-	S_WAIT,			"Waiting in queue",
-	S_DNS,			"Looking up host",
-	S_CONN,			"Making connection",
-	S_SENT,			"Request sent",
-	S_GETH,			"Getting headers",
-	S_PROC,			"Server is processing request",
-	S_TRANS,		"Transferring",
+	S_WAIT,			TEXT(T_WAITING_IN_QUEUE),
+	S_DNS,			TEXT(T_LOOKING_UP_HOST),
+	S_CONN,			TEXT(T_MAKING_CONNECTION),
+	S_SENT,			TEXT(T_REQUEST_SENT),
+	S_GETH,			TEXT(T_GETTING_HEADERS),
+	S_PROC,			TEXT(T_SERVER_IS_PROCESSING_REQUEST),
+	S_TRANS,		TEXT(T_TRANSFERRING),
 
-	S_WAIT_REDIR,		"Waiting for redirect confirmation",
-	S_OK,			"OK",
-	S_INTERRUPTED,		"Interrupted",
-	S_EXCEPT,		"Socket exception",
-	S_INTERNAL,		"Internal error",
-	S_OUT_OF_MEM,		"Out of memory",
-	S_NO_DNS,		"Host not found",
-	S_CANT_WRITE,		"Error writing to socket",
-	S_CANT_READ,		"Error reading from socket",
-	S_MODIFIED,		"Data modified",
-	S_BAD_URL,		"Bad URL syntax",
-	S_TIMEOUT,		"Receive timeout",
-	S_RESTART,		"Request must be restarted",
+	S_WAIT_REDIR,		TEXT(T_WAITING_FOR_REDIRECT_CONFIRMATION),
+	S_OK,			TEXT(T_OK),
+	S_INTERRUPTED,		TEXT(T_INTERRUPTED),
+	S_EXCEPT,		TEXT(T_SOCKET_EXCEPTION),
+	S_INTERNAL,		TEXT(T_INTERNAL_ERROR),
+	S_OUT_OF_MEM,		TEXT(T_OUT_OF_MEMORY),
+	S_NO_DNS,		TEXT(T_HOST_NOT_FOUND),
+	S_CANT_WRITE,		TEXT(T_ERROR_WRITING_TO_SOCKET),
+	S_CANT_READ,		TEXT(T_ERROR_READING_FROM_SOCKET),
+	S_MODIFIED,		TEXT(T_DATA_MODIFIED),
+	S_BAD_URL,		TEXT(T_BAD_URL_SYNTAX),
+	S_TIMEOUT,		TEXT(T_RECEIVE_TIMEOUT),
+	S_RESTART,		TEXT(T_REQUEST_MUST_BE_RESTARTED),
+	S_STATE,		TEXT(T_CANT_GET_SOCKET_STATE),
 
-	S_HTTP_ERROR,		"Bad HTTP response",
-	S_HTTP_100,		"HTTP 100 (???)",
-	S_HTTP_204,		"No content",
+	S_HTTP_ERROR,		TEXT(T_BAD_HTTP_RESPONSE),
+	S_HTTP_100,		TEXT(T_HTTP_100),
+	S_HTTP_204,		TEXT(T_NO_CONTENT),
 
-	S_FILE_TYPE,		"Unknown file type",
-	S_FILE_ERROR,		"Error opening file",
+	S_FILE_TYPE,		TEXT(T_UNKNOWN_FILE_TYPE),
+	S_FILE_ERROR,		TEXT(T_ERROR_OPENING_FILE),
 
-	S_FTP_ERROR,		"Bad FTP response",
-	S_FTP_UNAVAIL,		"FTP service unavailable",
-	S_FTP_LOGIN,		"Bad FTP login",
-	S_FTP_PORT,		"FTP PORT command failed",
-	S_FTP_NO_FILE,		"File not found",
-	S_FTP_FILE_ERROR,	"FTP file error",
+	S_FTP_ERROR,		TEXT(T_BAD_FTP_RESPONSE),
+	S_FTP_UNAVAIL,		TEXT(T_FTP_SERVICE_UNAVAILABLE),
+	S_FTP_LOGIN,		TEXT(T_BAD_FTP_LOGIN),
+	S_FTP_PORT,		TEXT(T_FTP_PORT_COMMAND_FAILED),
+	S_FTP_NO_FILE,		TEXT(T_FILE_NOT_FOUND),
+	S_FTP_FILE_ERROR,	TEXT(T_FTP_FILE_ERROR),
 
 	0,		NULL
 };
