@@ -5,7 +5,10 @@ int hard_write(int fd, unsigned char *p, int l)
 	int w = 1;
 	int t = 0;
 	while (l > 0 && w) {
-		if ((w = write(fd, p, l)) < 0) return -1;
+		if ((w = write(fd, p, l)) < 0) {
+			if (errno == EINTR) continue;
+			return -1;
+		}
 		t += w;
 		p += w;
 		l -= w;
@@ -13,18 +16,55 @@ int hard_write(int fd, unsigned char *p, int l)
 	return t;
 }
 
+int hard_read(int fd, unsigned char *p, int l)
+{
+	int r = 1;
+	int t = 0;
+	while (l > 0 && r) {
+		if ((r = read(fd, p, l)) < 0) {
+			if (errno == EINTR) continue;
+			return -1;
+		}
+		/*{int ww;for(ww=0;ww<r;ww++)fprintf(stderr," %02x",(int)p[ww]);fflush(stderr);}*/
+		t += r;
+		p += r;
+		l -= r;
+	}
+	return t;
+}
+
+unsigned char *get_cwd()
+{
+	int bufsize = 128;
+	unsigned char *buf;
+	while (1) {
+		if (!(buf = mem_alloc(bufsize))) return NULL;
+		if (getcwd(buf, bufsize)) return buf;
+		mem_free(buf);
+		if (errno == EINTR) continue;
+		if (errno != ERANGE) return NULL;
+		bufsize += 128;
+	}
+	return NULL;
+}
+
+void set_cwd(unsigned char *path)
+{
+	if (path) while (chdir(path) && errno == EINTR) ;
+}
+
 struct list_head terminals = {&terminals, &terminals};
 
 void alloc_term_screen(struct terminal *term, int x, int y)
 {
-	int *s, *t;
-	if ((s = mem_realloc(term->screen, x * y * sizeof(int)))) {
-		if ((t = mem_realloc(term->last_screen, x * y * sizeof(int)))) {
-			memset(t, -1, x * y * sizeof(int));
+	unsigned *s, *t;
+	if ((s = mem_realloc(term->screen, x * y * sizeof(unsigned)))) {
+		if ((t = mem_realloc(term->last_screen, x * y * sizeof(unsigned)))) {
+			memset(t, -1, x * y * sizeof(unsigned));
 			term->x = x;
 			term->y = y;
 			term->last_screen = t;
-			memset(s, 0, x * y * sizeof(int));
+			memset(s, 0, x * y * sizeof(unsigned));
 			term->screen = s;
 			term->dirty = 1;
 		}
@@ -38,7 +78,7 @@ void check_if_no_terminal();
 void clear_terminal(struct terminal *term)
 {
 	fill_area(term, 0, 0, term->x, term->y, ' ');
-	set_cursor(term, 0, 0);
+	set_cursor(term, 0, 0, 0, 0);
 }
 
 void redraw_terminal_ev(struct terminal *term, int e)
@@ -66,7 +106,11 @@ void redraw_terminal_all(struct terminal *term)
 
 void erase_screen(struct terminal *term)
 {
-	if (term->blocked == -1) hard_write(term->fdout, "\033[2J\033[1;1H", 10);
+	if (!term->master || !is_blocked()) {
+		if (term->master) want_draw();
+		hard_write(term->fdout, "\033[2J\033[1;1H", 10);
+		if (term->master) done_draw();
+	}
 }
 
 void redraw_terminal_cls(struct terminal *term)
@@ -76,6 +120,12 @@ void redraw_terminal_cls(struct terminal *term)
 	redraw_terminal_all(term);
 }
 
+void cls_redraw_all_terminals()
+{
+	struct terminal *term;
+	foreach(term, terminals) redraw_terminal_cls(term);
+}
+
 void redraw_from_window(struct window *win)
 {
 	struct terminal *term = win->term;
@@ -144,8 +194,9 @@ void delete_window(struct window *win)
 void delete_window_ev(struct window *win, struct event *ev)
 {
 	struct window *w = win->next;
+	if ((void *)w == &win->term->windows) w = NULL;
 	delete_window(win);
-	if (ev && w->next != w) w->handler(w, ev, 1);
+	if (ev && w && w->next != w) w->handler(w, ev, 1);
 }
 
 void set_window_ptr(struct window *win, int x, int y)
@@ -215,6 +266,41 @@ void add_empty_window(struct terminal *term, void (*fn)(void *), void *data)
 	add_window(term, empty_window_handler, ewd);
 }
 
+void free_term_specs()
+{
+	free_list(term_specs);
+}
+
+struct list_head term_specs = {&term_specs, &term_specs};
+
+struct term_spec dumb_term = { NULL, NULL, "", 0, 0, 0, 0, 0 };
+
+struct term_spec *get_term_spec(unsigned char *term)
+{
+	struct term_spec *t;
+	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
+	return &dumb_term;
+}
+
+struct term_spec *new_term_spec(unsigned char *term)
+{
+	struct term_spec *t;
+	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
+	if (!(t = mem_alloc(sizeof(struct term_spec)))) return NULL;
+	memcpy(t, &dumb_term, sizeof(struct term_spec));
+	if (strlen(term) < MAX_TERM_LEN) strcpy(t->term, term);
+	else memcpy(t->term, term, MAX_TERM_LEN - 1), t->term[MAX_TERM_LEN - 1] = 0;
+	add_to_list(term_specs, t);
+	sync_term_specs();
+	return t;
+}
+
+void sync_term_specs()
+{
+	struct terminal *term;
+	foreach (term, terminals) term->spec = get_term_spec(term->term);
+}
+
 struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct window *, struct event *, int))
 {
 	struct terminal *term;
@@ -226,6 +312,7 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 	memset(term, 0, sizeof(struct terminal));
 	term->fdin = fdin;
 	term->fdout = fdout;
+	term->master = term->fdout == get_output_handle();
 	/*term->x = 0;
 	term->y = 0;
 	term->cx = 0;
@@ -237,7 +324,9 @@ struct terminal *init_term(int fdin, int fdout, void (*root_window)(struct windo
 	term->blocked = -1;
 	term->screen = DUMMY;
 	term->last_screen = DUMMY;
+	term->spec = &dumb_term;
 	term->term[0] = 0;
+	term->cwd[0] = 0;
 	term->input_queue = DUMMY;
 	term->qlen = 0;
 	init_list(term->windows);
@@ -266,8 +355,8 @@ void in_term(struct terminal *term)
 		return;
 	}
 	term->input_queue = iq;
-	if ((r = read(term->fdin, iq + term->qlen, ALLOC_GR)) == -1) {
-		error("ERROR: error on terminal: could not read event");
+	if ((r = read(term->fdin, iq + term->qlen, ALLOC_GR)) <= 0) {
+		if (r == -1 && errno != ECONNRESET) error("ERROR: error %d on terminal: could not read event", errno);
 		destroy_terminal(term);
 		return;
 	}
@@ -282,17 +371,25 @@ void in_term(struct terminal *term)
 	}
 	if (ev->ev == EV_INIT) {
 		int init_len;
-		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + 2 * sizeof(int)) return;
-		init_len = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + sizeof(int));
-		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + 2 * sizeof(int) + init_len) return;
+		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int)) return;
+		init_len = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		if (term->qlen < sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len) return;
 		memcpy(term->term, iq + sizeof(struct event), MAX_TERM_LEN);
 		term->term[MAX_TERM_LEN - 1] = 0;
-		term->xwin = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN);
-		ev->b = (long)(iq + sizeof(struct event) + MAX_TERM_LEN + sizeof(int));
-		r = sizeof(struct event) + MAX_TERM_LEN + 2 * sizeof(int) + init_len;
+		memcpy(term->cwd, iq + sizeof(struct event) + MAX_TERM_LEN, MAX_CWD_LEN);
+		term->cwd[MAX_CWD_LEN - 1] = 0;
+		term->environment = *(int *)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN);
+		ev->b = (long)(iq + sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + sizeof(int));
+		r = sizeof(struct event) + MAX_TERM_LEN + MAX_CWD_LEN + 2 * sizeof(int) + init_len;
+		sync_term_specs();
 	}
 	if (ev->ev == EV_REDRAW || ev->ev == EV_RESIZE || ev->ev == EV_INIT) {
 		struct window *win;
+		send_redraw:
+		if (ev->x < 0 || ev->y < 0) {
+			error("ERROR: bad terminal size: %d, %d", (int)ev->x, (int)ev->y);
+			goto mm;
+		}
 		alloc_term_screen(term, ev->x, ev->y);
 		clear_terminal(term);
 		erase_screen(term);
@@ -300,7 +397,16 @@ void in_term(struct terminal *term)
 		foreachback(win, term->windows) win->handler(win, ev, 0);
 		term->redrawing = 0;
 	}
-	if (ev->ev == EV_KBD || ev->ev == EV_MOUSE) ((struct window *)&term->windows)->next->handler(term->windows.next, ev, 0);
+	if (ev->ev == EV_KBD || ev->ev == EV_MOUSE) {
+		if (ev->ev == EV_KBD && upcase(ev->x) == 'L' && ev->y == KBD_CTRL) {
+			ev->ev = EV_REDRAW;
+			ev->x = term->x;
+			ev->y = term->y;
+			goto send_redraw;
+		}
+		else if (ev->ev == EV_KBD && ev->x == KBD_CTRL_C) ((struct window *)&term->windows)->prev->handler(term->windows.prev, ev, 0);
+		else ((struct window *)&term->windows)->next->handler(term->windows.next, ev, 0);
+	}
 	if (ev->ev == EV_ABORT) destroy_terminal(term);
 	/*redraw_screen(term);*/
 	mm:
@@ -316,6 +422,14 @@ inline int getcompcode(int c)
 
 unsigned char frame_dumb[48] =	"   ||||++||++++++--|-+||++--|-+----++++++++     ";
 unsigned char frame_vt100[48] =	"aaaxuuukkuxkjjjkmvwtqnttmlvwtqnvvwwmmllnnjla    ";
+unsigned char frame_koi[48] = {
+	144,145,146,129,135,178,180,167,
+	166,181,161,168,174,173,172,131,
+	132,137,136,134,128,138,175,176,
+	171,165,187,184,177,160,190,185,
+	186,182,183,170,169,162,164,189,
+	188,133,130,141,140,142,143,139,
+};
 unsigned char frame_restrict[48] = {
 	0, 0, 0, 0, 0, 179, 186, 186,
 	205, 0, 0, 0, 0, 186, 205, 0,
@@ -346,6 +460,7 @@ unsigned char frame_restrict[48] = {
 			else add_to_str(&a, &l, "\x0e");				\
 		}									\
 		if (mode && c >= 176 && c < 224) c = frame_vt100[c - 176];		\
+	} else if (s->mode == TERM_KOI8 && (ch >> 15) && c >= 176 && c < 224) { c = frame_koi[c - 176];\
 	} else if (s->mode == TERM_DUMB && (ch >> 15) && c >= 176 && c < 224) c = frame_dumb[c - 176];\
 	if (!(A & 0100) && (A >> 3) == (A & 7)) A = (A & 070) | 7 * !(A & 020);		\
 	if (A != attrib) {								\
@@ -385,9 +500,9 @@ void redraw_screen(struct terminal *term)
 	int mode = -1;
 	int l = 0;
 	struct term_spec *s;
-	if (!term->dirty || term->blocked != -1) return;
+	if (!term->dirty || (term->master && is_blocked())) return;
 	if (!(a = init_str())) return;
-	s = get_term_spec(term->term);
+	s = term->spec;
 	for (y = 0; y < term->y; y++)
 		for (x = 0; x < term->x; x++, p++) {
 			if (y == term->y - 1 && x == term->x - 1) break;
@@ -422,7 +537,9 @@ void redraw_screen(struct terminal *term)
 		add_num_to_str(&a, &l, term->cx + 1);
 		add_to_str(&a, &l, "H");
 	}
+	if (l && term->master) want_draw();
 	hard_write(term->fdout, a, l);
+	if (l && term->master) done_draw();
 	mem_free(a);
 	memcpy(term->last_screen, term->screen, term->x * term->y * sizeof(int));
 	term->dirty = 0;
@@ -431,6 +548,8 @@ void redraw_screen(struct terminal *term)
 void destroy_terminal(struct terminal *term)
 {
 	while ((term->windows.next) != &term->windows) delete_window(term->windows.next);
+	/*if (term->cwd) mem_free(term->cwd);*/
+	if (term->title) mem_free(term->title);
 	mem_free(term->screen);
 	mem_free(term->last_screen);
 	set_handlers(term->fdin, NULL, NULL, NULL, NULL);
@@ -440,8 +559,20 @@ void destroy_terminal(struct terminal *term)
 		set_handlers(term->blocked, NULL, NULL, NULL, NULL);
 	}
 	del_from_list(term);
+	close(term->fdin);
+	if (term->fdout != 1) {
+		if (term->fdout != term->fdin) close(term->fdout);
+	} else {
+		unhandle_terminal_signals(term);
+		free_all_itrms();
+#ifndef NO_FORK_ON_EXIT
+		if (!list_empty(terminals)) {
+			if (fork()) exit(0);
+		}
+#endif
+	}
 	mem_free(term);
-	check_if_no_terminal(); /* !!! FIXME */
+	check_if_no_terminal();
 }
 
 void destroy_all_terminals()
@@ -452,38 +583,10 @@ void destroy_all_terminals()
 
 void check_if_no_terminal()
 {
-	if (terminals.next != &terminals) return;
+	if (!list_empty(terminals)) return;
 	terminate = 1;
 }
 
-void free_term_specs()
-{
-	free_list(term_specs);
-}
-
-struct list_head term_specs = {&term_specs, &term_specs};
-
-struct term_spec dumb_term = { NULL, NULL, "", 0, 0, 0, 0, 0 };
-
-struct term_spec *get_term_spec(unsigned char *term)
-{
-	struct term_spec *t;
-	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
-	return &dumb_term;
-}
-
-struct term_spec *new_term_spec(unsigned char *term)
-{
-	struct term_spec *t;
-	foreach(t, term_specs) if (!strcasecmp(t->term, term)) return t;
-	if (!(t = mem_alloc(sizeof(struct term_spec)))) return NULL;
-	memcpy(t, &dumb_term, sizeof(struct term_spec));
-	if (strlen(term) < MAX_TERM_LEN) strcpy(t->term, term);
-	else memcpy(t->term, term, MAX_TERM_LEN - 1), t->term[MAX_TERM_LEN - 1] = 0;
-	add_to_list(term_specs, t);
-	return t;
-}
-
 void set_char(struct terminal *t, int x, int y, unsigned c)
 {
 	t->dirty = 1;
@@ -524,7 +627,7 @@ void set_line_color(struct terminal *t, int x, int y, int l, unsigned c)
 	int i;
 	t->dirty = 1;
 	for (i = x >= 0 ? 0 : -x; i < (x+l <= t->x ? l : t->x-x); i++)
-		t->screen[x+i + t->x * y] = (t->screen[x+i + t->x * y] & 0x80fff) | (c & ~0x80ff);
+		t->screen[x+i + t->x * y] = (t->screen[x+i + t->x * y] & 0x80ff) | (c & ~0x80ff);
 }
 
 void fill_area(struct terminal *t, int x, int y, int xw, int yw, unsigned c)
@@ -558,9 +661,10 @@ void print_text(struct terminal *t, int x, int y, int l, unsigned char *text, un
 	for (; l-- && *text; text++, x++) set_char(t, x, y, *text + c);
 }
 
-void set_cursor(struct terminal *term, int x, int y)
+void set_cursor(struct terminal *term, int x, int y, int altx, int alty)
 {
 	term->dirty = 1;
+	if (term->spec->block_cursor) x = altx, y = alty;
 	if (x >= term->x) x = term->x - 1;
 	if (y >= term->y) y = term->y - 1;
 	if (x < 0) x = 0;
@@ -571,9 +675,12 @@ void set_cursor(struct terminal *term, int x, int y)
 
 void exec_thread(unsigned char *path, int p)
 {
-	exe(path);
+#if defined(HAVE_SETPGID) && !defined(BEOS) && !defined(HAVE_BEGINTHREAD)
+	if (path[0] == 2) setpgid(0, 0);
+#endif
+	exe(path + 1);
 	close(p);
-	if (path[strlen(path) + 1]) unlink(path + strlen(path) + 1);
+	if (path[1 + strlen(path + 1) + 1]) unlink(path + 1 + strlen(path + 1) + 1);
 }
 
 void close_handle(void *p)
@@ -594,38 +701,73 @@ void unblock_terminal(struct terminal *term)
 
 void exec_on_terminal(struct terminal *term, unsigned char *path, unsigned char *delete, int fg)
 {
+	if (path && !*path) return;
+	if (!path) path="";
 #ifdef NO_FG_EXEC
 	fg = 0;
 #endif
-	if (fg && term->blocked != -1) {
-		unlink(delete);
-		return;
-	}
-	if (term->fdout == 1) {
-		int blockh;
-		unsigned char *param;
-		if (!(param = mem_alloc(strlen(path) + strlen(delete) + 2))) return;
-		strcpy(param, path);
-		strcpy(param + strlen(path) + 1, delete);
-		if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, strlen(path) + strlen(delete) + 2)) == -1) {
+	if (term->master) {
+		if (!*path) dispatch_special(delete);
+		else {
+			int blockh;
+			unsigned char *param;
+			if (is_blocked() && fg) {
+				unlink(delete);
+				return;
+			}
+			if (!(param = mem_alloc(strlen(path) + strlen(delete) + 3))) return;
+			param[0] = fg;
+			strcpy(param + 1, path);
+			strcpy(param + 1 + strlen(path) + 1, delete);
+			if (fg == 1) block_itrm(term->fdin);
+			if ((blockh = start_thread((void (*)(void *, int))exec_thread, param, strlen(path) + strlen(delete) + 3)) == -1) {
+				if (fg == 1) unblock_itrm(term->fdin);
+				mem_free(param);
+				return;
+			}
 			mem_free(param);
-			return;
-		}
-		mem_free(param);
-		if (fg) {
-			term->blocked = blockh;
-			set_handlers(blockh, (void (*)(void *))unblock_terminal, NULL, (void (*)(void *))unblock_terminal, term);
-			set_handlers(term->fdin, NULL, NULL, (void (*)(void *))destroy_terminal, term);
-			block_itrm(term->fdin);
-		} else {
-			set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
+			if (fg == 1) {
+				term->blocked = blockh;
+				set_handlers(blockh, (void (*)(void *))unblock_terminal, NULL, (void (*)(void *))unblock_terminal, term);
+				set_handlers(term->fdin, NULL, NULL, (void (*)(void *))destroy_terminal, term);
+				/*block_itrm(term->fdin);*/
+			} else {
+				set_handlers(blockh, close_handle, NULL, close_handle, (void *)blockh);
+			}
 		}
 	} else {
-		char x = 0;
+		unsigned char *data;
+		if ((data = mem_alloc(strlen(path) + strlen(delete) + 4))) {
+			data[0] = 0;
+			data[1] = fg;
+			strcpy(data + 2, path);
+			strcpy(data + 3 + strlen(path), delete);
+			hard_write(term->fdout, data, strlen(path) + strlen(delete) + 4);
+			mem_free(data);
+		}
+		/*char x = 0;
 		hard_write(term->fdout, &x, 1);
 		x = fg;
 		hard_write(term->fdout, &x, 1);
 		hard_write(term->fdout, path, strlen(path) + 1);
-		hard_write(term->fdout, path, strlen(delete) + 1);
+		hard_write(term->fdout, delete, strlen(delete) + 1);*/
 	}
 }
+
+void do_terminal_function(struct terminal *term, unsigned char code, unsigned char *data)
+{
+	unsigned char *x_data;
+	if (!(x_data = mem_alloc(strlen(data) + 2))) return;
+	x_data[0] = code;
+	strcpy(x_data + 1, data);
+	exec_on_terminal(term, NULL, x_data, 0);
+	mem_free(x_data);
+}
+
+void set_terminal_title(struct terminal *term, unsigned char *title)
+{
+	if (term->title && !strcmp(title, term->title)) return;
+	if (term->title) mem_free(term->title);
+	term->title = stracpy(title);
+	do_terminal_function(term, TERM_FN_TITLE, title);
+}
