@@ -4,39 +4,40 @@ struct {
 	unsigned char *prot;
 	int port;
 	void (*func)(struct connection *);
-	void (*nc_func)(unsigned char *);
+	void (*nc_func)(struct session *, unsigned char *);
 	int free_syntax;
-	int file_prot;
-} protocols[]=	{
-		{"file", 0, file_func, NULL, 1, 1},
-		{"http", 80, http_func, NULL, 0, 1},
-		{"proxy", 3128, proxy_func, NULL, 0, 1},
-		{"ftp", 21, ftp_func, NULL, 0, 1},
-/*		{"mailto", 0, NULL, mailto_func, 1, 0},*/
-/*		{"telnet", 0, NULL, telnet_func, 0, 0},*/
-		{NULL, 0, NULL}};
+	int need_slashes;
+	int need_slash_after_host;
+} protocols[]= {
+		{"file", 0, file_func, NULL, 1, 1, 0},
+		{"http", 80, http_func, NULL, 0, 1, 1},
+		{"proxy", 3128, proxy_func, NULL, 0, 1, 1},
+		{"ftp", 21, ftp_func, NULL, 0, 1, 1},
+		{"finger", 79, finger_func, NULL, 0, 1, 1},
+		{"mailto", 0, NULL, mailto_func, 0, 0, 0},
+		{"telnet", 0, NULL, telnet_func, 0, 0, 0},
+		{"tn3270", 0, NULL, tn3270_func, 0, 0, 0},
+		{NULL, 0, NULL}
+};
 
 int check_protocol(unsigned char *p, int l)
 {
-	unsigned char *pn;
 	int i;
-	if (!(pn = memacpy(p, l))) return -1;
 	for (i = 0; protocols[i].prot; i++)
-		if (!strcasecmp(protocols[i].prot, pn)) {
-			mem_free(pn);
-			return protocols[i].free_syntax;
+		if (!casecmp(protocols[i].prot, p, l)) {
+			return i;
 		}
-	mem_free(pn);
 	return -1;
 }
 
-int get_prot_info(unsigned char *prot, int *port, void (**func)(struct connection *))
+int get_prot_info(unsigned char *prot, int *port, void (**func)(struct connection *), void (**nc_func)(struct session *ses, unsigned char *))
 {
 	int i;
 	for (i = 0; protocols[i].prot; i++)
 		if (!strcasecmp(protocols[i].prot, prot)) {
 			if (port) *port = protocols[i].port;
 			if (func) *func = protocols[i].func;
+			if (nc_func) *nc_func = protocols[i].nc_func;
 			return 0;
 		}
 	return -1;
@@ -60,17 +61,20 @@ int parse_url(unsigned char *url, int *prlen, unsigned char **user, int *uslen,
 	if (dalen) *dalen = 0;
 	if (post) *post = NULL;
 	if (!url || !(p = strchr(url, ':'))) return -1;
-	if (p[1] != '/' || p[2] != '/') return -1;
 	if (prlen) *prlen = p - url;
 	if ((a = check_protocol(url, p - url)) == -1) return -1;
-	if (a) {
+	if (p[1] != '/' || p[2] != '/') {
+		if (protocols[a].need_slashes) return -1;
+		p -= 2;
+	}
+	if (protocols[a].free_syntax) {
 		if (data) *data = p + 3;
 		if (dalen) *dalen = strlen(p + 3);
 		return 0;
 	}
 	p += 3;
 	q = p + strcspn(p, "@/");
-	if (!*q) return -1;
+	if (!*q && protocols[a].need_slash_after_host) return -1;
 	if (*q == '@') {
 		unsigned char *pp = strchr(p, ':');
 		if (!pp || pp > q) {
@@ -85,6 +89,7 @@ int parse_url(unsigned char *url, int *prlen, unsigned char **user, int *uslen,
 		p = q + 1;
 	} 
 	q = p + strcspn(p, ":/");
+	if (!*q && protocols[a].need_slash_after_host) return -1;
 	if (host) *host = p;
 	if (holen) *holen = q - p;
 	if (*q == ':') {
@@ -158,7 +163,7 @@ int get_port(unsigned char *url)
 		if (n) return n;
 	}
 	if ((h = get_protocol_name(url))) {
-		get_prot_info(h, &n, NULL);
+		get_prot_info(h, &n, NULL, NULL);
 		mem_free(h);
 	}
 	return n;
@@ -169,7 +174,17 @@ void (*get_protocol_handle(unsigned char *url))(struct connection *)
 	unsigned char *p;
 	void (*f)(struct connection *) = NULL;
 	if (!(p = get_protocol_name(url))) return NULL;
-	get_prot_info(p, NULL, &f);
+	get_prot_info(p, NULL, &f, NULL);
+	mem_free(p);
+	return f;
+}
+
+void (*get_external_protocol_function(unsigned char *url))(struct session *, unsigned char *)
+{
+	unsigned char *p;
+	void (*f)(struct session *, unsigned char *) = NULL;
+	if (!(p = get_protocol_name(url))) return NULL;
+	get_prot_info(p, NULL, NULL, &f);
 	mem_free(p);
 	return f;
 }
@@ -244,11 +259,9 @@ void translate_directories(unsigned char *url)
 				goto b;
 			}
 		}
-		d = d1;
-		goto p;
+		/*d = d1;
+		goto p;*/
 		b:
-		/**d++ = s[0];*/
-		/*d++;*/
 		s += 3;
 		goto r;
 	}
@@ -256,7 +269,23 @@ void translate_directories(unsigned char *url)
 	if ((*d++ = *s++)) goto r;
 }
 
-/* !!! FIXME: "proxy://" */
+void insert_wd(unsigned char **up, unsigned char *cwd)
+{
+	unsigned char *url = *up;
+	if (!url || !cwd || !*cwd) return;
+	if (casecmp(url, "file://", 7)) return;
+	if (dir_sep(url[7])) return;
+#ifdef DOS_FS
+	if (upcase(url[7]) >= 'A' && upcase(url[7]) <= 'Z' && url[8] == ':' && dir_sep(url[9])) return;
+#endif
+	if (!(url = mem_alloc(strlen(*up) + strlen(cwd) + 2))) return;
+	memcpy(url, *up, 7);
+	strcpy(url + 7, cwd);
+	if (!dir_sep(cwd[strlen(cwd) - 1])) strcat(url, "/");
+	strcat(url, *up + 7);
+	mem_free(*up);
+	*up = url;
+}
 
 unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 {
@@ -281,6 +310,7 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 		add_to_strn(&n, rel);
 		return n;
 	}
+	if (!casecmp("proxy://", rel, 8)) goto prx;
 	if (!parse_url(rel, &l, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
 		n = stracpy(rel);
 		translate_directories(n);
@@ -295,6 +325,7 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 		}
 		mem_free(n);
 	}
+	prx:
 	if (parse_url(base, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &p, NULL, NULL) || !p) {
 		internal("bad base url");
 		return NULL;
@@ -311,57 +342,66 @@ unsigned char *join_urls(unsigned char *base, unsigned char *rel)
 	return n;
 }
 
-unsigned char *translate_url(unsigned char *url)
+unsigned char *translate_url(unsigned char *url, unsigned char *cwd)
 {
 	static int r = 0;
 	unsigned char *ch;
 	unsigned char *nu, *da;
+	while(*url == ' ') url++;
+	if (!casecmp("proxy://", url, 8)) goto prx;
 	if (!parse_url(url, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &da, NULL, NULL)) {
-		if (!r && !casecmp(url, "proxy://", 8)) {
-			unsigned char *p;
-			r = 1;
-			p = translate_url(da);
-			r = 0;
-			if (!p) return NULL;
-			nu = memacpy(url, da - url);
-			add_to_strn(&nu, p);
-			mem_free(p);
-		} else {
-			nu = stracpy(url);
-			translate_directories(nu);
-		}
+		nu = stracpy(url);
+		insert_wd(&nu, cwd);
+		translate_directories(nu);
 		return nu;
 	}
 	if (strstr(url, "//") && (nu = stracpy(url))) {
 		add_to_strn(&nu, "/");
 		if (!parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+			insert_wd(&nu, cwd);
 			translate_directories(nu);
 			return nu;
 		}
 		mem_free(nu);
 	}
+	prx:
 	ch = url + strcspn(url, ".:/@");
 	if (*ch != ':' || *(url + strcspn(url, "/@")) == '@') {
 		unsigned char *prefix = "file://";
 		int sl = 0;
-		if (*url != '.' && *ch == '.') prefix = "http://", sl = 1;
+		if (*url != '.' && *ch == '.') {
+			unsigned char *f, *e;
+			int i;
+			for (e = ch + 1; *(f = e + strcspn(e, ".:/")) == '.'; e = f + 1) ;
+			for (i = 0; i < f - e; i++) if (e[i] >= '0' && e[i] <= '9') goto http;
+			if (f - e == 2) http: prefix = "http://", sl = 1;
+			else if (f - e == 3) {
+				unsigned char *tld[] = { "com", "edu", "net", "org", "gov", "mil", "int", NULL };
+				for (i = 0; tld[i]; i++) if (!casecmp(tld[i], e, 3)) goto http;
+			}
+		}
 		if (*ch == '@' || *ch == ':' || !cmpbeg(url, "ftp.")) prefix = "ftp://", sl = 1;
 		if (!(nu = stracpy(prefix))) return NULL;
 		add_to_strn(&nu, url);
 		if (sl && !strchr(url, '/')) add_to_strn(&nu, "/");
 		if (parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) mem_free(nu), nu = NULL;
-		else translate_directories(nu);
+		else {
+			insert_wd(&nu, cwd);
+			translate_directories(nu);
+		}
 		return nu;
 	}
 	if (!(nu = memacpy(url, ch - url + 1))) return NULL;
 	add_to_strn(&nu, "//");
 	add_to_strn(&nu, ch + 1);
 	if (!parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+		insert_wd(&nu, cwd);
 		translate_directories(nu);
 		return nu;
 	}
 	add_to_strn(&nu, "/");
 	if (!parse_url(nu, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL)) {
+		insert_wd(&nu, cwd);
 		translate_directories(nu);
 		return nu;
 	}
